/*
 * Goldwater Distributed Web Architecture.
 * - Goldwater Halt Command
 *
 * Copyright 1996-2002 netFluid Technology Ltd
 *
 * Goldwater is written and maintained by netFluid Technology Ltd.
 * http://www.nfluid.co.uk
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 *
 * $Revision: 1.4 $  $Date: 2003/09/13 20:30:24 $
 * 
 * --------------------------------------------------------------------------
 * $Log: gw_halt_cmd.c,v $
 * Revision 1.4  2003/09/13 20:30:24  csmith
 * Better error message when service failes to execute.  Flush boot message so text up to -> appears whilst server is booting so user can see what is being waited for (this should always have been the case, but the flush was omitted for some reason :o)
 *
 * Revision 1.3  2003/09/13 13:04:32  csmith
 * Portable.net PInvoke helper functions.
 * Guardian bug: Did not clean BBReqQueue table on failure/shutdown of servers.
 * Increased version numbers to reflect significant bug fixes.
 * Announce copyright FSF in administrator.
 * Administrator now pauses 1/10 second between boot/halt commands to give
 * guardian time to breath (not strictly necessary, but may avoid some signal
 * race conditions).
 *
 * Revision 1.2  2003/03/29 12:21:10  csmith
 * cvs merge to 0.3.4
 *
 * Revision 1.9  2003/03/26 19:22:35  chris
 * Added cycle command.
 *
 * Revision 1.8  2002/11/30 14:04:42  chris
 * GPL headers
 *
 * Revision 1.7  2002/06/10 17:34:47  chris
 * All phlib includes have phlib/ prefix and gwlib includes have nothing.
 *
 * Revision 1.6  2002/05/21 21:15:16  chris
 * Big tidy up - lots of misc changes.
 * Changed 'internal' services to be prefixed by '!' so they don't show in
 * the admin tool service list.
 *
 * Revision 1.5  2002/05/12 13:15:45  chris
 * Cleanup of unnecessary fields in BBEntry.
 *
 * Revision 1.4  2002/03/14 20:45:55  chris
 * Decent server halting!
 *
 * Revision 1.3  2002/03/12 22:50:59  chris
 * Added BBReqQueue support.  Currently defines a BBReqQueue entry for
 * each server, even though servers are grouped into classes and thus
 * will usually share a queue.
 *
 * Revision 1.2  2002/03/11 22:34:30  chris
 * More uptodate TODO and other things.
 *
 * Revision 1.1  2002/03/10 15:30:39  chris
 * New support
 *
 * --------------------------------------------------------------------------
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <phlib/xl_types.h>
#include <phlib/xl_string.h>
#include <phlib/xl_linklst.h>

#include <gwreq.h>
#include <gwlog.h>
#include <gwclient.h>
#include <gwapi.h>

/* Privates */
#include "goldreq.h"
#include "regstrings.h"
#include "gwreadconf.h"
#include "gwgbprm.h"
#include "mutex.h"

#include "gwadmin.h"

typedef struct {
  long classid;
  long srvid;
  char server[SRV_NM_MAX + 1];
} SrvDet;

#define INV_CLASS  "ERR: Invalid Class Identifier\n"
#define INV_SERVER "ERR: Invalid Server Identifier\n"
#define HALT_USAGE "Usage: halt -s server|-c class id -i service id|-c class id\
n\n"

static RetValue halt_server( long class, long instance );
static RetValue compile_halt_list( long classid, long srvid, LIST *hlist );

RetValue kill_server( BootParams *bp );

extern char *local_dom;
extern char  local_dom_lev[];

/* ------------------------------------------------------------------------- */

RetValue
gw_halt( CString *argCS )
{
  long  val       = 0;
  CString optvCS = CSInit;
  char    opt    = '0';
  int     go     = 1;
  long    class  = 0;
  long    id     = 0;
  char    server[128] = {0};


  while( CSLenPtr( argCS ) > 0 ) {
    
    if( parse_arguments( argCS, &optvCS, &opt, "s:i:c:a" ) == Failure ) {
	  /* Give Usage */
	  printf( HALT_USAGE );
	  go = 0;
	  break;
	}

    switch( opt ) {
    case 'c': /* Class */
      if( CStoInt32( &optvCS, &val ) == Failure ) {
	  printf( INV_CLASS ); /* Error */
	  go = 0;
      } else class = val;
      break;
      
    case 'i': /* ID */
      if( CStoInt32( &optvCS, &val ) == Failure ) {
	  printf( INV_SERVER );
	  go = 0;
      } else id = val;
      break;
      
    case 's': /* Server */
      if( CSLen( optvCS ) > sizeof(server) - 1 ) 
	CSLen( optvCS ) = sizeof(server) - 1;
      
      strncpy( server, CSStart(optvCS), CSLen(optvCS) );
      server[ CSLen(optvCS) ] = '\0';
      break;

    case 'a': /* Everything */
      break;

    default: 
      printf( HALT_USAGE );
      go = 0;
      break;
    }
  }

  if( id >= 0 && class == -1 && server[0] == '\0' ) {
    printf( INV_CLASS );
	go = 0;
  }

  if( go ) halt_server( class, id );
  
  return Success;
}

/* ------------------------------------------------------------------------- */

static RetValue
halt_server( long class, long instance )
{
  int     sct    = 0;
  int     fct    = 0;
  SrvDet *sd     = {0};
  LIST   *hlist;

  BootParams bp;

  printf( "\nHalting server processes...\n\n" );
  
  xl_List_init( &hlist, sizeof( SrvDet ) );

  /* Compile a list of servers to halt from BB in one hit...
   */
  compile_halt_list( class, instance, hlist );
  
  if( xl_List_get_first( hlist, (uchar_t**)&sd ) == Success ) {
    do {
      memset( (void*)&bp, 0, sizeof(BootParams) );

      bp.sp.servclass = sd->classid;
      bp.sp.servid    = sd->srvid;
      strcpy( bp.sp.server, sd->server );
      (kill_server( &bp ) == Success) ? sct++ : fct++;
      
    } while( xl_List_get_next( hlist, (uchar_t**)&sd ) == Success );
  } /* endwhile */

  xl_List_end( &hlist );

  printf( "\nHalted %d servers, %d failed\n", sct, fct );
  printf( "\n" );

  return Success;
}

/* ------------------------------------------------------------------------- */

static RetValue
compile_halt_list( long classid, long srvid, LIST *hlist )
{
  RetValue ret = Success;
  uchar   *ptr = NULL;
  SrvDet   sd  = {0};
  SHMLIST *BB  = (SHMLIST*)((char*)bbc + bbc->BB_entry_offset );
  
  /* Get Mutex */
  if( aquire_bbc_mutex( MUTEX_BBC ) == Failure ) {
    gw_logf( LOG_SERIOUS, "Could not aquire BB Mutex" );
    return Failure;
  }

  /*
   * Search BB for this server
   */
  if( shm_List_get_first( BB, &ptr ) == Success ) {
    do {
      if( ((BBEntry*)(ptr))->domainid > bbc->domainid ) break;
      
      if( ((BBEntry*)(ptr))->domainid == bbc->domainid ) {

	/* SKIP OVER ADMINISTRATIVE CLASS!!!!! - DON'T WANT TO KILL GUARDIAN!!
	 */
	if( ((BBEntry*)(ptr))->bb_sp.servclass == 1 ) continue;
	
	/* If we want a particular class, and we've overshot, then finish */
	if( classid && ((BBEntry*)(ptr))->bb_sp.servclass > classid ) break;

	/* If we want any class, or a particular class... */
	if( !classid || ((BBEntry*)(ptr))->bb_sp.servclass == classid ) {      
	  /* If we want any server instance, or a particular one... */
	  if( !srvid || srvid == ((BBEntry*)(ptr))->bb_sp.servid ) {

	    sd.classid = ((BBEntry*)(ptr))->bb_sp.servclass;
	    sd.srvid   = ((BBEntry*)(ptr))->bb_sp.servid;
	    strcpy( sd.server,((BBEntry*)(ptr))->bb_sp.server );

	    if( xl_List_insert_node( hlist, ListEnd, 
				     (uchar_t*)&sd ) == Failure ) {
	      ret = Failure;
	      break;
	    }
	  }
	}
      }
    } while( shm_List_get_next( BB, &ptr ) == Success );
  }
  
  /* Release Mutex */
  if( release_bbc_mutex( MUTEX_BBC ) == Failure ) {
    gw_logf( LOG_SERIOUS, "Could not release BB Mutex" );
    return Failure;    
  }
  
  return ret;
}

/* ------------------------------------------------------------------------- */

RetValue
kill_server( BootParams *bp )
{
  GWMessage   msg  = {0}; 
  int         ret  = 0;
  RetValue    lret = Failure;
  BootParams *rbp  = NULL;

  printf( "  ++ Halting %-15.15s Class: %-6ld ID: %-6ld -> ",
	  bp->sp.server, bp->sp.servclass, bp->sp.servid );
  fflush(NULL);

  ret = gw_svc_call( "!/gctrl/srvhalt", (char*)bp, sizeof(BootParams), &msg, 0 );

  rbp = (BootParams *)msg.data;

  if( ret >= 0 ) {
    printf( "Pid: %-ld ... %.*s\n",rbp->pid,(int)rbp->resp_len,rbp->resp_text);
    lret = Success;
  } else {
    
    if( gwerror == GWE_APPERR ) {
      printf( "Fail: %.*s\n", (int)rbp->resp_len, rbp->resp_text );
    } else {
      printf( "Fail: %s\n", gw_errorstr( gwerror ) );
    }
  }

  gw_free_message( &msg );

  usleep( 100000 );

  return Success;
}

/* ------------------------------------------------------------------------- */
/* end */
