/*
 * Domain Bridge Message Functions
 *
 * Copyright 2002 netFluid Technology Ltd
 *
 * Goldwater is written and maintained by netFluid Technology Ltd.
 * http://www.nfluid.co.uk
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * --------------------------------------------------------------------------
 * $Id: dbcom.c,v 1.2 2003/03/29 12:21:10 csmith Exp $
 * --------------------------------------------------------------------------
 */
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include <phlib/xl_types.h>
#include <phlib/xl_memory.h>

#include <goldreq.h>
#include <gwlog.h>
#include <gwgbprm.h>

#include "dbridge.h"
#include "dbctrl.h"
#include "dbcom.h"


/* ------------------------------------------------------------------------- */

RetValue
gdbm_create_header( GDBM_Version ver, DomainMessage *dm, 
                    int tdomid, DomSession *sc )
{
  int min_head_size   = 0;
  unsigned char *hbuf = NULL;
  BBDom *domarray     = NULL;

  domarray = (BBDom*)((char *)bbc + bbc->BB_dom_offset);

  switch( ver ) {
  case DBM_VERSION_001: 
    min_head_size = DBM_PAYLOAD_IX;
  break;

  default:
    gwerror = GWE_BADDMVER;
	return Failure;
  }

  reset_cstring( &sc->headCS ); /* Empty header buf, retain memory */

  /* Make sure headCS contains enough space to be going on with */
  if( CSLen(sc->headCS) < min_head_size ) {
    if( extend_cstring( &sc->headCS, min_head_size ) == Failure ) {
      gwerror = GWE_SYSFAIL;
	  return Failure;
	}
  }

  hbuf = CSStart( sc->headCS ); /* Get ptr to buffer */

  /* Set length of CString to be the length of our header */
  CSLen( sc->headCS ) = min_head_size;

  /* Build header */

  /* Copy 128 bit msgid and txid into header in byte order
   */
  memcpy( (void*)(hbuf + DBM_MESSAGE_ID_IX), (void*)dm->rctrl.msgid, sizeof(uuid_t) );
  memcpy( (void*)(hbuf + DBM_TXID_IX),       (void*)dm->txid,  sizeof(uuid_t) );

  /* Now the rest of the header
   */
  store_int32_LittleEndian( hbuf + DBM_MAGIC_NUM_IX, (int32)DBM_MAGIC_NUMBER  );
  store_int32_LittleEndian( hbuf + DBM_VERSION_IX,   (int32)ver               );
  store_int32_LittleEndian( hbuf + DBM_MSG_TYPE_IX,  (int32)dm->rctrl.msgtype );
  store_int32_LittleEndian( hbuf + DBM_MSG_REP_ID_IX,(int32)dm->rctrl.org_r_msgqid );

  strncpy( hbuf + DBM_SERVICE_IX, dm->rctrl.service, SVC_NM_MAX );
  strncpy( hbuf + DBM_FROM_IX, _Domain, DOM_NM_MAX );

  /* *** MUTEX *** TODO: */
  strncpy( hbuf + DBM_TO_IX,   domarray[tdomid-1].domain ,DOM_NM_MAX );
  /* *** MUTEX *** TODO: */

  store_int32_LittleEndian( hbuf + DBM_FLAGS_IX,     (int32)dm->rctrl.flags   );
  store_int32_LittleEndian( hbuf + DBM_PRIORITY_IX,  (int32)1                 );
  store_int32_LittleEndian( hbuf + DBM_DATA_LEN_IX,  (int32)CSLen(dm->rctrl.cache) );

  store_int32_LittleEndian( hbuf + DBM_TIMEOUT_IX,   (int32)_ReqTout      );

  store_int32_LittleEndian( hbuf + DBM_AUTH_TYPE_IX, (int32)DBM_AUTH_MD5  );
  store_int32_LittleEndian( hbuf + DBM_AUTH_LEN_IX,  (int32)16            );

#if 0
  {
    #include <sys/types.h>
    #include <fcntl.h>
    int fd = open( "/tmp/domdump", O_WRONLY | O_CREAT | O_TRUNC, 0666 );
	write( fd, hbuf, min_head_size );
	close(fd);
  }
#endif

  return Success;
}

/* ------------------------------------------------------------------------- */

int
decode_ack( DomSession *sc, unsigned char *buf )
{
  sc->status.result = read_int32_LittleEndian( buf + DBR_STATUS_IX    );
  sc->status.error  = read_int32_LittleEndian( buf + DBR_SYS_ERROR_IX );
  
  return sc->status.result;
}

/* ------------------------------------------------------------------------- */
/* end */
