/*
 * Goldwater Distributed Web Architecture.
 *
 * Server Guardian
 *
 * Copyright 1996-2003 netFluid Technilogy Limited
 *
 * THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF NETFLUID TECHNOLOGY LIMITED
 *
 * The copyright notice above does not evidence any actual or intended 
 * publication of such source code.    
 *
 * $Revision: 1.11 $
 *
 * ---------------------------------------------------------------------------
 *
 * $Log: guardian.c,v $
 * Revision 1.11  2003/10/23 20:51:56  csmith
 * Contains compile time support for valgrind (This is for internal debugging)
 *
 * Revision 1.10  2003/09/16 20:13:33  csmith
 * rationalise ipc usage into gwipc.c
 *
 * Revision 1.9  2003/09/14 19:02:21  csmith
 * Slimmed child_signal a bit
 *
 * Revision 1.8  2003/09/14 18:50:58  csmith
 * Process Guardian is now more pro-active with respect to detecting the state
 * of the child server processes.  It no longer blocks on the message queue
 * indefinately but instead wakes up every 5 seconds and does a waitpid for
 * any child termination signals that may have been generated.  Appropriate
 * action is then taken.
 * Previously child signals were just checked by trapping SIGCHLD, but there
 * are times (under very heavy load) that signals get lost.  The new scanning
 * mechanism ensures that they always get picked up.
 *
 * Revision 1.7  2003/09/13 20:30:07  csmith
 * Better error message when service failes to execute.  
 * Flush boot message so text up to -> appears whilst server is booting so user 
 * can see what is being waited for (this should always have been 
 * the case, but the flush was omitted for some reason :o)
 *
 * Revision 1.6  2003/09/13 14:08:26  csmith
 * Some tidying up
 *
 * Revision 1.5  2003/09/13 13:04:32  csmith
 * Portable.net PInvoke helper functions.
 * Guardian bug: Did not clean BBReqQueue table on failure/shutdown of servers.
 * Increased version numbers to reflect significant bug fixes.
 * Announce copyright FSF in administrator.
 * Administrator now pauses 1/10 second between boot/halt commands to give
 * guardian time to breath (not strictly necessary, but may avoid some signal
 * race conditions).
 *
 * Revision 1.4  2003/05/29 22:45:07  csmith
 * Write FSF Copyright out to logfile when starting up.
 *
 * Revision 1.3  2003/04/12 18:09:54  csmith
 * SIGCONT didn't work under freeBSD, but SIGHUP did. Hmm.
 *
 * Revision 1.2  2003/03/29 12:21:10  csmith
 * cvs merge to 0.3.4
 *
 * Revision 1.66  2003/03/25 21:44:56  chris
 * Fixed up version numbers for next stable release.
 *
 * Revision 1.65  2003/03/25 21:16:22  chris
 * Enable auto-recycle of servers
 *
 * Revision 1.64  2003/01/05 22:58:34  chris
 * Use an implicit path to the system servers (in location like /usr/local/bin)
 * This avoids having to use execvp and thus the potential for a ruid root
 * program searching the path for an exe.
 *
 * Revision 1.63  2003/01/04 16:28:47  chris
 * gw_init is now intelligent about detecting if Goldwater is running.
 * By repeatedly calling gw_init() a client can make sure they stay attached
 * without having to do gw_init(), gw_release() over and over.  gw_init() will
 * detect that the backend has been bounced and try its hardest to reconnect.
 *
 * Revision 1.62  2003/01/04 13:24:19  chris
 * Removed #if 1
 *
 * Revision 1.61  2003/01/03 00:09:09  chris
 * Use internal umask when creating files, shm and queues
 *
 * Revision 1.60  2002/12/28 00:24:47  chris
 * Through use of lsof discovered files (log and stdout/stderr) that were open
 * multiple times.  Close them before exec'ing servers.
 *
 * Revision 1.59  2002/12/27 16:48:41  chris
 * Some comments where 'env' would be extracted from gwconfig
 *
 * Revision 1.58  2002/09/01 11:24:30  chris
 * On shutdown of Goldwater, give Guardian time to die before ripping IPC
 * to shreds.
 *
 * Revision 1.57  2002/08/27 18:55:56  chris
 * Do not try and compile services for system servers - this prevents silly
 * warnings about the missing service entries in the config.
 *
 * Revision 1.56  2002/08/27 17:30:15  chris
 * Don't hold BBC_MUTEX over call to launch_server in restart_*.
 *
 * Revision 1.55  2002/08/17 13:50:23  chris
 * Use gwsys_num_cpus from gwlib
 *
 * Revision 1.54  2002/07/30 21:32:16  chris
 * Check for APPDIR existance *before* creating a msg queue, because if APPDIR
 * fails we end up with stale queues.
 *
 * Revision 1.53  2002/06/17 23:44:07  chris
 * Comments.
 *
 * Revision 1.52  2002/06/17 23:40:22  chris
 * Make sure guardian is a session leader and thus does not have a
 * controlling tty.
 *
 * Revision 1.51  2002/06/17 19:25:30  chris
 * New server GWNServer requires the same exec_server treatment as GWDBridge
 * (search $PATH for it).
 *
 * Revision 1.50  2002/06/10 17:34:47  chris
 * All phlib includes have phlib/ prefix and gwlib includes have nothing.
 *
 * Revision 1.49  2002/06/09 20:20:43  chris
 * Make sure gw_log identifier is set as Guardian when entering signal
 * handler.
 *
 * Revision 1.48  2002/06/04 09:34:48  chris
 * Cannot pass -1 as msgqid in server parameters as the -1 is taken as an option!
 *
 * Revision 1.47  2002/05/26 18:23:30  chris
 * Pass GWS_EXPORTABLE to __gw_add_service if service is exported.
 *
 * Revision 1.46  2002/05/21 21:15:16  chris
 * Big tidy up - lots of misc changes.
 * Changed 'internal' services to be prefixed by '!' so they don't show in
 * the admin tool service list.
 *
 * Revision 1.45  2002/05/14 21:43:53  chris
 * Boot GWDBridge as a system server (so scan $PATH instead of implicit
 * path).
 *
 * Revision 1.44  2002/05/14 20:20:27  chris
 * Moved service add/removal functions into gwlib.
 * Made sure msgid used for net_server is -1 so that we don't ever try and
 * remove it (the first msgqid ever created could have the qid of 0).
 *
 * Revision 1.43  2002/05/14 19:10:44  chris
 * Add server entry to BB before it is actually booted so that the BB entry
 * is available from the moment the server is started.  It is inserted with
 * a false PID until the exec_server call returns and the PID property of
 * the BB entry can be updated with the real value.
 *
 * Revision 1.42  2002/05/12 13:15:48  chris
 * Cleanup of unnecessary fields in BBEntry.
 *
 * Revision 1.41  2002/05/12 11:18:32  chris
 * Block signals when in handling services - a sigchld or child_oaky_sig
 * may cause a mutex to be held 'claimed' and a deadlock to occur.
 *
 * Revision 1.40  2002/04/21 20:23:00  chris
 * Setup local netserver address and port into globals GWHost and GWPort
 * Logic surrounding compile_services was reversed (wrt netserver)
 *
 * Revision 1.39  2002/04/14 19:30:00  chris
 * Detection of a netserver was done on srvclass.  Changed this to be on
 * server name (exe name) so any class id may be assgned by the administrator.
 *
 * Revision 1.38  2002/04/14 00:33:50  chris
 * Allow booting of netserver (which is *slightly* different from normal
 * servers in that it does not need a message queue and does not have any
 * service).
 * remove_queue modified to only try and remove queues > 0
 *
 * Revision 1.37  2002/04/12 20:13:51  chris
 * Just some comments.
 *
 * Revision 1.36  2002/03/17 17:53:26  chris
 * In Guardians server exit callback, clean the BB.
 *
 * Revision 1.35  2002/03/17 13:49:01  chris
 * Fixed errors in scanning config for our classid for services.
 *
 * Revision 1.34  2002/03/17 11:54:02  chris
 * New service name handling and fixed bugs wrt passing domainid and not
 * domain name through functions.
 *
 * Revision 1.33  2002/03/16 17:29:16  chris
 * Debugging for compile_services.
 *
 * Revision 1.32  2002/03/12 22:51:01  chris
 * Added BBReqQueue support.  Currently defines a BBReqQueue entry for
 * each server, even though servers are grouped into classes and thus
 * will usually share a queue.
 *
 * Revision 1.31  2002/03/11 19:34:43  chris
 * Cleaned up Guardian and made sure the halting of services removed BB entries.
 *
 * Revision 1.30  2002/03/10 19:35:30  chris
 * Handles proper homed messages (initially for shutdown of servers).
 *
 * Revision 1.29  2002/03/04 00:23:18  chris
 * Cleaned massages.
 *
 * Revision 1.28  2002/03/02 17:02:23  chris
 * Return BootParam block as reply from boot request.
 *
 * Revision 1.27  2002/02/24 19:36:28  chris
 * The new server and service addition code was not claiming/releasing
 * mutexs properly.
 * Cleaned up warnings..
 *
 * Revision 1.26  2002/02/24 18:52:10  chris
 * Changed name of boot service to .GCTRLBOOT and removed some logging.
 *
 * Revision 1.25  2002/02/24 18:10:49  chris
 * Massive Change: Guardian now starts as a normal GWServer.  It bootstraps
 * itself into the BB and adds it's own services to the Service Table before
 * settling down and waiting for service requests (like a normal server).
 * It now uses the compiled version of the XML based configuration file.
 *
 * Revision 1.24  2002/02/16 16:41:56  chris
 * Introduced autoconf
 *
 * Revision 1.23  2002/01/03 20:27:31  chris
 * Introduction of preliminary Domain Control handling.
 *
 * Revision 1.22  2001/12/26 22:57:42  chris
 * Changed all references to Imagepage Computing to netFluid Technology.
 *
 * Revision 1.21  2001/11/02 22:18:07  chris
 * Give port/host in log error when failing to connect to registry.
 *
 * Revision 1.20  2001/10/18 00:04:44  chris
 * After forking a child server, if the exec fails then DO NOT call exit (3)
 * because this will invoke the ExitCleanup function registered with atexit()
 * and remove all the queues!
 * Call _exit() instead as this does not invoke any registered exit functions.
 *
 * Revision 1.19  2001/04/17 17:12:07  chris
 * In event of service table becoming full, act as though the boot was
 * successful (the server will be running and in the server table) but
 * warn that service table is full and that server may not be accessable.
 *
 * Revision 1.18  2001/04/17 13:38:31  chris
 * Uses domain id when searching for servers.
 *
 * Revision 1.17  2001/04/15 13:00:09  chris
 * Introduced domain table.
 *
 * Revision 1.16  2001/04/12 15:28:57  chris
 * Passes ip (numeric dot format) and port to compile_services().  These
 * values complete the service info that may be exported to other
 * application servers.
 *
 * Revision 1.15  2001/04/01 23:30:52  chris
 * Startup log now at SYSINFO.
 *
 * Revision 1.14  2001/04/01 23:25:08  chris
 * Added version.h and startup log message.
 *
 * Revision 1.13  2001/04/01 23:11:02  chris
 * Whenever we try to remove a server from the BB we remove any
 * associated reply queue.  We also perform a check to see if there are
 * any other servers of the same class present, and if not, we remove the
 * request queue as well.  Added Guardian Shutdown command.
 *
 * Revision 1.12  2001/03/30 18:44:33  chris
 * Rationalised logging.
 *
 * Revision 1.11  2001/03/26 18:37:32  chris
 * Handles controlled termination of servers via messages.
 *
 * Revision 1.10  2001/03/26 14:35:25  chris
 * Moved BB searching over to gwlib (bbquery.c)
 *
 * Revision 1.9  2001/03/26 14:00:52  chris
 * When booting servers, orders SHM on class and id.
 *
 * ---------------------------------------------------------------------------
 */
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <setjmp.h>
#include <signal.h>
#include <sys/wait.h>
#include <errno.h>
#include <sys/msg.h>

#define GW_INTERNAL_FLAGS 1  /* get internal flags - Hidden for user apps */

#include "version.h"

#include <phlib/xl_types.h>
#include <phlib/xl_malloc.h>
#include <phlib/xl_logging.h>
#include <phlib/xl_socket.h>

#include <goldreq.h>
#include <gwipc.h>

#include <mutex.h>
#include <gwreadconf.h>
#include <gwapi.h>
#include <gwpaths.h>

#define GW_PRV_CLIENT 1
# include <gwclient.h>
#undef  GW_PRV_CLIENT

#include <gwgbprm.h>
#include <gwlog.h>
#include <gwmsg.h>
#include <bbquery.h>
#include <gwservice.h>

#include <regstrings.h>
#include <regdata.h>

#include <gwdomain.h>
#include <gwsystem.h>

#include <gwuuid.h>

#define STARTUP_MESSAGE1 "Guardian Version " VERSION " Ready"
#define STARTUP_MESSAGE2 "(C)1999-2003 netFluid Technology Ltd"
#define STARTUP_MESSAGE3 "(C)2003 Free Software Foundation, Inc"

static const char ONETSERVER[] = "gwnet";
static const char SYSSERVER1[] = "GWDBridge";
static const char SYSSERVER2[] = "GWNServer";

#define SIGJMP_OKAY    0x00
#define CHILD_READY    0x01
#define CHILD_FAILURE  0x02


#define COULD_NOT_EXEC 123

#define DO_NOT_RESPAWN  0
#define ALLOW_RESPAWN   1
#define CONTROLLED_SHUT 2

#define PRE_START_PID   999999

#define MAX_SANE_RESTART 50 /* Don't restart a server more than n times ! */

static long   guardian_queue_idx = -1;

static long   _start_timer  = 5;

static int    child_error   = 0;
static int    child_sig     = 0;

static pid_t  booting_child_pid = 0;
static sigjmp_buf jbuf;

static char  *resp_text     = "";
static int    boot_error    = 0;

BBDom *local_dom     = NULL;
char   phoenix_host[64];
long   _gwadmin_port = 0;
long   _reg_port     = 0;

/* Prototypes
 */
static RetValue add_server_entry    ( BBEntry *,long,long,
                                      long,long *,long,pid_t, BBEntry** );
static RetValue exec_server         (const char *,const char *,
                                     int,int,int,pid_t *);
static RetValue launch_single_server( BootParams * );
static RetValue shutdown_server     ( BootParams *bp, pid_t *pid );
static RetValue restart_child       ( pid_t pid );
static RetValue clean_bb_for_server ( pid_t pid );
static void     child_signal        ( int sig   );
static long     create_queue_entry  ( long );
static void     delete_queue_entry  ( long qidx );
static void     ExitCleanup         ( );
static void     QueueCleanup        ( );
static RetValue compile_services    ( const char *server, BBEntry *bbe, 
                                      SHMLIST *BBS, const char *domain );

/* ------------------------------------------------------------------------- */
/*
 * Special Server Initialisation for Guardian Process.
 *
 * The Guardian has to add itself to the BB, including Domain, Server and
 * service info.
 *
 * Once it has done this, it may tell it's parent that it has started
 * and may enter the normal request wait of traditional servers.
 *
 * - The techniques used here ARE NOT TO BE COPIED by developers of
 *   Goldwater Servers.  Please use the provided API instead of poking
 *   about in the BB.  The API is not usable until Guardian has got up
 *   and running, which is why it's not used here.
 */
GWStatus 
GWSRV_INIT( char *s_name, char *reserved )
{
  struct sigaction sa, sao;
  char *msg  = NULL;
  int   nc   = 0;
  long  qidx = 0;
  BBEntry  bbe; 
  BBEntry *nbbe_p = NULL; 
  char uuidbuf[64];

  /* We're the master controlling process.  Make sure we've not got a 
   * controlling terminal.
   */
  setsid();

  gw_log_init( _ServId, "Guardian" );

  nc = gwsys_num_cpus();
  gw_logf( LOG_SYSINFO, STARTUP_MESSAGE1 );
  gw_logf( LOG_SYSINFO, STARTUP_MESSAGE2 );
  gw_logf( LOG_SYSINFO, STARTUP_MESSAGE3 );
  gw_logf( LOG_SYSINFO, "Detected %d Processor%s", nc, (nc==1) ? "" : "s" );

  if( gw_config_init() < 0 ) {
  
    switch(gwerror) {
    case GWE_CFGRERR: msg = "Cannot open configuration\n";          break;
    case GWE_BADCFG : msg = "Goldwater configuration is invalid\n"; break;
    case GWE_BCFGARC: msg = "Configuration not for this architecture\n"; break;
    }
    gw_logf( LOG_SERIOUS, msg );
    
    return GWFailure;
  }

  /*
   * Grab semaphore so Goldwater can check if Guardian is running
   */
  if( aquire_bbc_mutex( MUTEX_GUARDIAN ) == Failure ) {
    gw_logf( LOG_SERIOUS, "Cannot aquire guardian mutex" );
    return GWFailure;
  }

  gw_logf( LOG_INTDBG, "Real UID: %d  Effective UID: %d", getuid(), geteuid());

  /* Configure the local domain - We have to do this here because we're the
   * master process, and will be up before any system servers get booted
   * (after all, this process boots them!). This means that the DomainBridge
   * server cannot be given the responsibility!
   */
  if( (local_dom = __gw_domain_from_config( _Domain )) == NULL ) {
    gw_logf( LOG_SERIOUS, "Cannot initialise local domain %s", _Domain );
    return GWFailure;
  }
  bbc->domainid = local_dom->domainid;

  /* Set local domain to be Available */
  __gw_domain_set_state( _Domain, GWD_AVAILABLE );

  /* Set the local hosts netServer address and port */
  strcpy( bbc->gwhost, local_dom->gw_ipaddr );
  bbc->gwport = local_dom->gw_port;

  /* Add ourselves to the server table
   * - First set some default values in BBEntry
   */
  bbe.bb_sp.max = 1;
  bbe.bb_sp.min = 1;
  strncpy( bbe.bb_sp.server, s_name,  SRV_NM_MAX );
  
  qidx = -1; /* We've no BBReqQueue entry (yet) for the Guardian Queue! */

  if( add_server_entry( &bbe, local_dom->domainid, _ServClass, _ServId, 
			&qidx, bbc->guardian_msgqid, getpid(), &nbbe_p ) == Failure ) {

    gw_logf( LOG_SERIOUS, "Cannot initialise Server Table" );
    return GWFailure;
  }
  guardian_queue_idx = qidx;

  /* Advertise our internal services */
  if( gw_add_service( "!/gctrl/srvboot", "gctrl_boot" ) < 0 ) {
    gw_logf( LOG_SERIOUS, "Cannot add service entry gctrl_boot" );
    return GWFailure;
  }

  if( gw_add_service( "!/gctrl/srvhalt", "gctrl_halt" ) < 0 ) {
    gw_logf( LOG_SERIOUS, "Cannot add service entry gctrl_halt" );
    return GWFailure;
  }

  if( gw_add_service( "!/gctrl/shutdown", "gctrl_exit" ) < 0 ) {
    gw_logf( LOG_SERIOUS, "Cannot add service entry gctrl_exit" );
    return GWFailure;
  }

  /* Trap SIGCHLD and CHILD_OKAY_SIG
   */
  sa.sa_handler = child_signal; 
  sigemptyset( &sa.sa_mask );
  sa.sa_flags = SA_RESTART;
  sigaction( SIGCHLD, &sa, &sao );

  sa.sa_handler = child_signal;
  sigemptyset( &sa.sa_mask );
  sa.sa_flags = SA_RESTART; 
  sigaction( CHILD_OKAY_SIG, &sa, &sao );

  /* Generate a unique instance ID for the booted system
   * - Used by clients etc to verify that they're connected to the same
   *   instance they think they are.
   */
  uuid_generate( bbc->instance_id );
  uuid_unparse( bbc->instance_id, uuidbuf );
  gw_logf( LOG_SYSINFO, "Boot-Instance ID: %s", uuidbuf );

  return GWSuccess;
}

/* ------------------------------------------------------------------------- */

void
GWSRV_EXIT()
{
  BBEntry *bbe_p = NULL; 
  SHMLIST *BB    = (SHMLIST*)((char*)bbc + bbc->BB_entry_offset );

  /* When server exits clean up IPC Queues
   */
  ExitCleanup();

  /*
   * Search BB for Guardian
   */
  if( find_server_entry( local_dom->domainid, 
			 _ServClass, _ServId, &bbe_p, BB ) >= 0 ) {

    /* Clean BB for guardian */
    bbe_p->bb_sc.respawn = CONTROLLED_SHUT; 
    clean_bb_for_server( bbe_p->bb_sc.pid );
  }

  uuid_clear( bbc->instance_id );

  gw_config_deinit();
}

/* ------------------------------------------------------------------------- */
/* Special copy of server init code (srvinit.c) that does a timed wait for
 * requests, periodically waking up and checking on the state of it's children,
 * restarting them as necessary.
 *
 * Not to be duplicated by user applications as this hooks directly to the
 * internal API which will probably be subject to change.
 *
 * *** XXX You Have Been Warned XXX ***
 */
int
main( int argc, char **argv )
{
  int   done = 0;
  char *f    = 0;
  char *s    = 0;
  char *d    = 0;
  long fl    = 0;
  long ret   = 0;
  GWMsgCtrl *rmsg_ctrl = NULL;

  __gw_server_init( argc, argv );

  do {
    done = 0;

	/* Do a Timed block waiting for requests (Hardcoded in gwmsg.c to 5 secs)
	 */
    fl = GWP_REQTBLK; 
    ret = __gw_get_request( &f, &s, &d, &fl );

    /* If we have returned with a request message, deal with it
	 */
    rmsg_ctrl = __gw_get_requestCtrl(); /* Get pointer to request Ctrl block */
	if( ret >= 0 && !(rmsg_ctrl->cflags & GWC_NOREQM) ) {

      done = ( __gw_invoke_C_service() < 0 ) ? 0 : 1;
    }
    __gw_complete_request(done); /* done == 1 will update BB counters*/

	/* Now perform any housekeeping that we may need to do.
	 * For instance, checking on the state of our child processes...
	 */
	/* TODO Decide whether this is best done every wakup or every request?
	 * ie if GWC_NOREQM is set or not...
	 * At the moment it is for every request and every wakeup.
	 */
    child_signal( SIGCHLD ); /* Check for dead children */

  } while( 1 );

  exit(0);
  return 0;
}

/* ------------------------------------------------------------------------- */

static void
ExitCleanup()
{
  BBReqQueue *bbq = (BBReqQueue*)((char*)bbc + bbc->BB_queue_offset);

  if( release_bbc_mutex( MUTEX_GUARDIAN ) == Failure ) {
    gw_logf( LOG_SERIOUS, "Cannot release guardian mutex" );
  }

  QueueCleanup();

  /* Clear guardians queue entry
   */
  memset( (void*)&bbq[ guardian_queue_idx ], 0, sizeof(BBReqQueue));
  bbq[ guardian_queue_idx ].msgqid = -1;

  gw_log( LOG_SYSINFO|LOG_AUDIT, "Guardian shutdown", -1 );
}


/* ------------------------------------------------------------------------- */
  
void
gctrl_boot( char *svc, GWMessage *reqm )
{
  int    l         = 0;
  int    mlen      = 0;
  char  *params    = NULL;
  char  *reply     = NULL;
  RetValue ret     = Failure;
  sigset_t ss, sso;
  
  /* Block all signals - we don't want to be interrupted during a mutex loop
   */
  sigfillset ( &ss );
  sigprocmask( SIG_BLOCK, &ss, &sso );

  /* Force log title.... */
  gw_log_init( _ServId, "Guardian" );

  if( reqm->len < sizeof( BootParams ) ) {
    gw_logf( LOG_SERIOUS, "Message is not a BootParam block" );
    gw_send_reply( GWFailure, NULL, 0 );
    return;
  }

  params = (char*)reqm->data;

#if DEBUG
  gw_logf(LOG_INTDBG, "SrvCls: %ld",  ((BootParams*)params)->sp.servclass );
  gw_logf(LOG_INTDBG, "SrvID : %ld",  ((BootParams*)params)->sp.servid );
  gw_logf(LOG_INTDBG, "Max   : %ld",  ((BootParams*)params)->sp.max    );
  gw_logf(LOG_INTDBG, "Min   : %ld",  ((BootParams*)params)->sp.min    );
  gw_logf(LOG_INTDBG, "Thresh: %ld",  ((BootParams*)params)->sp.rstrt_thresh);
  gw_logf(LOG_INTDBG, "Domain: '%s'", ((BootParams*)params)->domain  );
  gw_logf(LOG_INTDBG, "Server: '%s'", ((BootParams*)params)->sp.server  );
#endif
  
  ret = launch_single_server( (BootParams *)params );
  
  if( ret == Failure )
    gw_logf( LOG_SERIOUS, "Server Boot Failure - %s", resp_text );

  l = strlen(resp_text);
  ((BootParams *)params)->resp_len = 0;
  
  /* Create buffer for BootParam block and response text */
  mlen = sizeof(BootParams) + l + 1;
  if( (reply = xl_malloc( mlen )) != NULL ) {
    
    ((BootParams*)params)->resp_len = l;
    memcpy( reply, params, sizeof(BootParams) );
    memcpy( ((BootParams*)reply)->resp_text, resp_text, l );
  }
  
  gw_send_reply(ret == Success ? GWSuccess : GWFailure, reply, reply ? mlen:0);

  /* Restore signals */
  sigprocmask( SIG_SETMASK, &sso, NULL );

  return;
}


/* ------------------------------------------------------------------------- */

void
gctrl_halt( char *svc, GWMessage *reqm )
{
  char *params = NULL;

  long  l     = 0;
  long  mlen  = 0;
  char *reply = NULL;
  pid_t pid   = 0;
  RetValue ret = Success;
  sigset_t ss, sso;

  /* Block all signals - we don't want to be interrupted during a mutex loop
   */
  sigfillset ( &ss );
  sigprocmask( SIG_BLOCK, &ss, &sso );
  
  /* Force log title.... */
  gw_log_init( _ServId, "Guardian" );

  if( reqm->len < sizeof( BootParams ) ) {
    gw_logf( LOG_SERIOUS, "Message is not a BootParam block" );
    gw_send_reply( GWFailure, NULL, 0 );
    return;
  }

  params = (char*)reqm->data;

#if DEBUG
  gw_logf(LOG_INTDBG, "SrvCls: %ld",  ((BootParams*)params)->sp.servclass );
  gw_logf(LOG_INTDBG, "SrvID : %ld",  ((BootParams*)params)->sp.servid );
  gw_logf(LOG_INTDBG, "Max   : %ld",  ((BootParams*)params)->sp.max    );
  gw_logf(LOG_INTDBG, "Min   : %ld",  ((BootParams*)params)->sp.min    );
  gw_logf(LOG_INTDBG, "Thresh: %ld",  ((BootParams*)params)->sp.rstrt_thresh);
  gw_logf(LOG_INTDBG, "Domain: '%s'", ((BootParams*)params)->domain  );
  gw_logf(LOG_INTDBG, "Server: '%s'", ((BootParams*)params)->sp.server  );
#endif
  
  ret = shutdown_server( (BootParams*)params, &pid );

  /* We don't know what the state of the server is in now - so Remove it's
   * BB entries anyway (regardless of return status)
   */
  clean_bb_for_server( pid );

  if( ret == Failure )
    gw_logf( LOG_SERIOUS, "Server Halt Failure - %s", resp_text );
  
  l = strlen(resp_text);
  ((BootParams*)params)->resp_len = 0;
  ((BootParams*)params)->pid      = pid;
  
  /* Create buffer for BootParam block and response text */
  mlen = sizeof(BootParams) + l + 1;
  if( (reply = xl_malloc( mlen )) != NULL ) {
    
    ((BootParams*)params)->resp_len = l;
    memcpy( reply, params, sizeof(BootParams) );
    memcpy( ((BootParams*)reply)->resp_text, resp_text, l );
  }

  gw_send_reply(ret == Success ? GWSuccess : GWFailure, reply, reply ? mlen:0);

  /* Restore signals */
  sigprocmask( SIG_SETMASK, &sso, NULL );
}

/* ------------------------------------------------------------------------- */
/* This service causes guardian to exit
 */
void
gctrl_exit( char *svc, GWMessage *reqm )
{
  int    l         = 0;
  int    mlen      = 0;
  char  *reply     = NULL;
  pid_t  gpid      = getpid();
  sigset_t ss, sso;
  
  /* Block all signals - we don't want to be interrupted during a mutex loop
   */
  sigfillset ( &ss );
  sigprocmask( SIG_BLOCK, &ss, &sso );
  
  /* Force log title.... */
  gw_log_init( _ServId, "Guardian" );

  resp_text = "Halted";
  l = strlen(resp_text);

  /* Create buffer for BootParam block and response text */
  mlen = sizeof(BootParams) + l + 1;
  if( (reply = xl_malloc( mlen )) != NULL ) {
    
    ((BootParams*)reply)->resp_len = l;
    memset( reply, 0, sizeof(BootParams) );
    memcpy( ((BootParams*)reply)->resp_text, resp_text, l );

    ((BootParams*)reply)->pid = gpid;
    ((BootParams*)reply)->resp_len = l;
  } else {
    mlen = 0;
    reply = "";
  }
  
  gw_send_reply( GWSuccess, reply, mlen );

  /* Restore signals */
  sigprocmask( SIG_SETMASK, &sso, NULL );

  exit(0); /* Exit Guardian */
}

/* ------------------------------------------------------------------------- */
/*
 * Queue Cleanup etc
 */
static void
QueueCleanup()
{
  struct sigaction sa, sao;
  uchar   *ptr = NULL;
  BBEntry *bbe_p; 

  SHMLIST    *BB  = (SHMLIST*)   ((char*)bbc + bbc->BB_entry_offset);
  BBReqQueue *bbq = (BBReqQueue*)((char*)bbc + bbc->BB_queue_offset);

  /*
   * We don't want to know about child terminations now!
   */
  sa.sa_handler = SIG_DFL; /* Can't set SIG_IGN according to POSIX */
  sigemptyset( &sa.sa_mask );
  sa.sa_flags = 0; 
  sigaction( SIGCHLD, &sa, &sao );
 
  /*
   * Search through shm looking for msgqid's owned by servers.
   * - remove these queues as we find them
   */

  /* Get Mutex */
  if( aquire_bbc_mutex( MUTEX_BBC ) == Failure ) {
    gw_logf( LOG_SERIOUS, "Cannot aquire mutex" );
    return;
  }

  if( shm_List_get_first( BB, &ptr ) == Success ) {
    do {
      bbe_p = (BBEntry*)ptr;

      /* Don't want to respawn servlet 
       * - not that it'll make a difference as we're not traping SIGCHLD
       *   anymore.  Makes the SHM a bit more consistent though.
       */
      bbe_p->bb_sc.respawn = DO_NOT_RESPAWN;

      /*
       * Don't bother trapping errors
       */   
   
      /* Request queue */
	  remove_queue( bbq[ bbe_p->req_queue_idx ].msgqid );
	  delete_queue_entry( bbe_p->req_queue_idx );

      /* Reply queue */
	  remove_queue( bbe_p->bb_sc.repmqid );

    } while( shm_List_get_next( BB, &ptr ) == Success ) ;
  }
  
  /* Release Mutex */
  if( release_bbc_mutex( MUTEX_BBC ) == Failure ) {
    gw_logf( LOG_SERIOUS, "Cannot release mutex" );
  }
}

/* ------------------------------------------------------------------------- */

static RetValue
launch_single_server( BootParams *bp )
{
  RetValue ret    = Success;
  char    *server = 0;
  char    *domain = 0;
  long     srvcls = 0;
  long     srvid  = 0;
  long     bbqidx = -1;
  pid_t    spid   = 0;
  int      msgqid = -1;
  int      new_q  = 0;
  int      old_net_server = 0;
  int      sys_server     = 0;
  int      existingqueue  = 0;

  BBEntry *bbe_p  = NULL; 
  BBEntry *nbbe_p = NULL; 
  BBEntry  bbe; 

  SHMLIST    *BB  = (SHMLIST*)   ((char*)bbc + bbc->BB_entry_offset );
  SHMLIST    *BBS = (SHMLIST*)   ((char*)bbc + bbc->BB_svc_offset   );
  BBReqQueue *bbq = (BBReqQueue*)((char*)bbc + bbc->BB_queue_offset );

  srvcls = ((BootParams*)bp)->sp.servclass;
  srvid  = ((BootParams*)bp)->sp.servid;
  server = ((BootParams*)bp)->sp.server;
  domain = ((BootParams*)bp)->domain;

  ((BootParams*)bp)->pid = 0;

  if( strcmp( server, ONETSERVER ) == 0 ) old_net_server = 1;
  if( strcmp( server, SYSSERVER1 ) == 0 ) sys_server = 1;
  if( strcmp( server, SYSSERVER2 ) == 0 ) sys_server = 1;

  /*
   * Search BB for this server.  If it exists then we've got a duplicate
   * server.
   */
  if( find_server_entry( local_dom->domainid, 
			 srvcls, srvid, &bbe_p, BB ) >= 0 ) {
    resp_text = "Duplicate server";
    boot_error    = 1;
    return Failure;
  }
  /* If we had a class match, then bbe_p will point to the last server match
   * within this class.  Use this to get the BBReqQueue index of the other
   * servers within this class....
   */
  bbqidx = -1;
  msgqid = -1;
  if( bbe_p ) {
    bbqidx = bbe_p->req_queue_idx; /* BBReqQueue index for this class */
    msgqid = bbq[ bbqidx ].msgqid; /* Request queue for this class    */
    existingqueue = 1;
  }

  /* NOT SURE ABOUT DOING CHDIR EVERY TIME
   *
   * The bad Appdir check MUST be done BEFORE creating a new queueu, just
   * in case it fails and leaves the queue in place!
   */
  if( chdir( bbc->appdir ) < 0 ) {
    resp_text = "Bad APPDIR";
    boot_error    = 1;
    gw_logf(LOG_SERIOUS,"Invalid APPDIR in shmem.  Please clear and restart.");
    return Failure;
  } 

  /* I really want to move this into create_queue_entry() 
   * - Problem - need msgqid *before* we can call add_server_entry Soooooo
   *   we probably *do* need to create the queue here. 
   *   Hmm, Hmm and double Hmm......
   */
  if( msgqid == -1 ) {
    /* We need to create a new queue for this server
     */
    if( old_net_server ) {
      msgqid = -1; /* old Netserver Does Not require a msgqueue */
    } else {
      if( (msgqid = get_private_queue() ) < 0 ) {
        resp_text = "QUEUE error";
        boot_error    = 1;
        return Failure;
      }
      new_q = 1;
    }
  }


  /* There is a need to add the server entry to the BB *before* the server is
   * booted so that their details are available from the moment they start up.
   * - This means that we can set MyServEntry in __gw_server_init and as such
   *   will allow gw api commands such as gw_add_service to be used straigt
   *   away.
   */
  /* 
   * Get a copy of the significant bootparam needed for SHM insertion,
   * and replace pid in BB Entry with real pid.
   */
  memcpy( &(bbe.bb_sp), &(bp->sp), sizeof( ServerParams) );

  /* Add Server Entry with FALSE pid 'PRE_START_PID'
   */
  ret = add_server_entry( &bbe, bbc->domainid, srvcls, srvid, 
                          &bbqidx, msgqid, PRE_START_PID, &nbbe_p );

  /* Stop this server from being RESPAWNABLE! 
   */
  nbbe_p->bb_sc.respawn = DO_NOT_RESPAWN; /* Disable Re-Spawn */
    
  if( ret == Failure ) {
    gw_logf( LOG_SERIOUS, "Could not add server %s to server table", server);
    resp_text  = "Cannot add to server table";
    boot_error = 1;

    if(!existingqueue) delete_queue_entry( bbqidx );
    return ret;
  }

  ret = exec_server( server, domain, msgqid, srvid, srvcls, &spid );

  if( ret == Success ) {
    resp_text = "Booted";
 
    ((BootParams*)bp)->pid = spid;
    nbbe_p->bb_sc.pid = spid;  /* Update pre-exec server entry with pid */
    nbbe_p->bb_sc.respawn = ALLOW_RESPAWN;

    /* Now we insert the service(s) into the BB (if we're not the
     * old_net_server! - it doesn't have services as it binds to a socket!!
     * or a sys_server)
     */
    if( ! (old_net_server || sys_server) ) {
      ret = compile_services( server, nbbe_p, BBS, local_dom->domain );
    }
      
    /* -- server has been added to SHM, so appears in server list and stats 
     *    So here we set a response string to show what has happened, and
     *    return Success to the caller because we HAVE booted the server!
     */
    if( ret == Failure ) {
      /* Compiling services failed..
       */
      boot_error = 1;
      gw_logf( LOG_SYSWARN, "%s - Server may not be accessable", resp_text );
 	
      ret = Success;
    }
  } else {

    /* Remove the entry in the BB that we created *before* the server was
     * exec'd - The PID is PRE_START_PID.
     */
    nbbe_p->bb_sc.respawn = CONTROLLED_SHUT; /* Allow clean_bb_.. to work! */
    clean_bb_for_server( PRE_START_PID );

    resp_text   = "Could not execute";

    if( child_sig != 0 ) {
      gw_logf( LOG_SERIOUS, "Server terminated by signal %d", child_sig );
      resp_text = "Server Terminated";
    }
    boot_error = 1;
  }

  return ret;
}

/* ------------------------------------------------------------------------- */

static RetValue
compile_services( const char *server, BBEntry *bbe, 
		  SHMLIST *BBS, const char *domain )
{
  RetValue sret  = Failure;
  int  ret       = 0;
  int  l         = 0;
  int  idx       = 0;
  int  id        = 0;
  long len       = 0;
  long pos       = 0;
  long class     = 0;
  long classnode = 0;
  char *ptr      = NULL;
  char *service  = "";
  char *func     = "";
  char *export   = "";
  char path[128] = {0};
  ServiceFlags sflags = 0;
  
  static char last_dom_name[ DOM_NM_MAX + 1 ] = {0};
  static long last_dom_idx = -1;

  /* Find the config index for the requested domain
   * (Servers in a single domain are usually booted together, so cache
   *  the config index for the last domain treated to avoid repeat
   *  lookups).
   */
  if( strcmp( last_dom_name, domain ) != 0 ) {
    
    strncpy( last_dom_name, domain, DOM_NM_MAX );
    
    idx = 0;
    do {
      sprintf( path, "%s%06d%s", R_DOMAIN_B, idx, R_DOM_NAME_L );
      
      ret = gw_config_read_char( path, &ptr, &len );
      
      if( ret >= 0 && strcmp( domain, ptr ) == 0 ) {
	id = idx;
	break;
      }
      
      idx ++;
    } while( ret >= 0 );
    
    if( id < 0 ) {
      gw_logf( LOG_SERIOUS, "Domain %s missing from gwconfig", domain );
      return Failure;
    }
    last_dom_idx = id;
  }
  
  /* Get Mutex */
  if( aquire_bbc_mutex( MUTEX_BBC_SVC ) == Failure ) {
    gw_logf( LOG_SERIOUS, "Could not aquire SVC Mutex" );
    return Failure;
  }
  
  /* Find Location of This Servers services within BB...
   */
  pos = __gw_find_service_entry( bbe->domainid, 
				 bbe->bb_sp.servclass, 
				 bbe->bb_sp.servid, 
				 BBS );
  
  /* find the services for our class
   */
  idx   = -1; /* Start from -1 as first pass through loop is a dummy run */
  class = -1;
  do {
    if( class == bbe->bb_sp.servclass ) break;

    idx++;

    /* Get class id */
    l = sprintf( path, "%s%06ld%s%06d%s", 
		 R_DOMAIN_B, last_dom_idx, 
		 R_SERVERS_SB, idx, R_ID_L );

    gw_logf( LOG_INTDBG, "Looking for our classid in: '%s'", path );

  } while( gw_config_read_long( path, &class ) >= 0 );


  do { /* This loop just allows us to break out and release mutex properly */

    if( class != bbe->bb_sp.servclass ) {
      gw_logf( LOG_SYSERR, "Services for class %d not found in config",
	       bbe->bb_sp.servclass );
      break;
    }
    
    /* Found services matching our class */
    classnode = idx;
    
    /* Zip through all services for this Domain/Class combo
     */
    idx = 0;
    sret = Success;
    do {
      l = sprintf( path, "%s%06ld%s%06ld%s%06d", 
		   R_DOMAIN_B, last_dom_idx, 
		   R_SERVERS_SB, classnode, 
		   R_SERVICES_SB, idx );
      
      /* First look for function name : MANDATORY */
      strcpy( path + l, R_SERVICE_L );
      gw_logf( LOG_INTDBG, "Looking up svc branch '%s'", path );
      
      if((ret = gw_config_read_char( path, &service, &len )) < 0 && idx == 0) {

        gw_logf( LOG_SYSERR, "Services for class %d not found in config",
                 bbe->bb_sp.servclass );
        sret = Failure;
      } else gw_logf( LOG_INTDBG, "Found Service '%s'", service );
      
      if( ret >= 0 ) {
	strcpy( path + l, R_FUNCTION_L );
	gw_logf( LOG_INTDBG, "Looking up function branch '%s'", path );

	if( gw_config_read_char( path, &func, &len ) < 0) func = (char*)server;
	
	gw_logf( LOG_INTDBG, "Adding service '%s', func '%s' to service table",
		 service, func );


        sflags = 0;

        /* See if service is exportable ? 
	 */
	strcpy( path + l, R_EXPORT_L );
	if( gw_config_read_char( path, &export, &len ) < 0 ) export = "";
	if( strcmp( export, "yes" ) == 0 ||
	    strcmp( export, "YES" ) == 0 ) sflags |= GWS_EXPORTABLE;

	if( sflags & GWS_EXPORTABLE ) 
          gw_logf( LOG_INTDBG, "Service %s is exported", service );

	
	/* Don't worry about failures here - errors get logged
	 */
	if( __gw_add_service( bbe->domainid, 
			      bbe->bb_sp.servclass,
			      bbe->bb_sp.servid,
			      bbe->req_queue_idx,
			      BBS, service, func, &pos, sflags ) == Failure ) {
	  
	  resp_text = "No room in service table";
	  ret = -1;
	}
      }
      
      idx++;
    } while( ret >= 0 );
    
  } while (0); /* allows us to get out and drop mutex */

  /* Release Mutex */
  if( release_bbc_mutex( MUTEX_BBC_SVC ) == Failure ) {
    gw_logf( LOG_SERIOUS, "Could not release SVC Mutex" );
    return Failure;
  }

  return sret;
}

/* ------------------------------------------------------------------------- */
/*
 * Adds a server entry to the BB (if there is room)
 *
 * If *bb_rqueue_idx is -VE then a new entry is created for the class of
 * servers and the idx of this entry is passed back.
 * A successful entry add will have the ptr to this BBEntry returned in *nbbe
 */
RetValue
add_server_entry( BBEntry *bbe, long domainid, long srvcls, long srvid, 
		  long *bb_rqueue_idx, long msgqid, pid_t pid, BBEntry **nbbe )
{
  RetValue ret  = Success;
  long     ipos = -1;
  long     qidx = -1;

  SHMLIST    *BB  = (SHMLIST*)   ((char*)bbc + bbc->BB_entry_offset );
  BBReqQueue *bbq = (BBReqQueue*)((char*)bbc + bbc->BB_queue_offset );

#if DEBUG
  gw_logf( LOG_ALWAYS, "Create server entry Class: %ld  ID: %ld  RQid: %ld", 
           srvcls, srvid, *bb_rqueue_idx );
#endif

  /* If we don't have the index of a request queue entry to attach to
   * then create a new one!
   */
  if( *bb_rqueue_idx < 0 ) {
    /* Create new Request Queue Control Segment
     */
    if( (qidx = create_queue_entry( msgqid )) < 0 ) {
      gw_logf( LOG_SERIOUS, "Cannot create Queue entry for server" );
      return Failure;
    }
    *bb_rqueue_idx = qidx;
  }

  /* Set up BB entry details for this new server:
   */
  bbq[ *bb_rqueue_idx ].numdeps++; /* Increase number of servers on queue */

  bbe->bb_sp.servid    = srvid;
  bbe->bb_sp.servclass = srvcls;

  bbe->bb_sc.gracerstrt_ind = 0;
  bbe->bb_sc.pid       = pid;
  bbe->bb_sc.reqcount  = 0;
  bbe->bb_sc.frstrt    = 0;
  bbe->bb_sc.respawn   = ALLOW_RESPAWN; /* Allow server to be respawned */
  bbe->bb_sc.repmqid   = -1;            /* No reply queue initially     */

  bbe->req_queue_idx   = *bb_rqueue_idx;
  bbe->domainid        = domainid;  
  
  strcpy( bbe->bb_sc.state, "Ready" );

  /*
   * (Order servers in SHM by class and serverid)
   */
  ipos = srch_for_class_id( domainid, srvcls, srvid, BB );
    
  /* Get Mutex */
  if( aquire_bbc_mutex( MUTEX_BBC ) == Failure ) {
    gw_logf( LOG_SERIOUS, "Cannot aquire mutex" );
    return Failure;
  }

  if( shm_List_insert_node( BB, (ipos < 0) ? ListEnd : ipos, 
			    (uchar*)bbe ) == Failure ) ret = Failure;

  if( ret == Success ) *nbbe = (BBEntry *)shm_List_get_current_data( BB );

  /* Release Mutex */
  if( release_bbc_mutex( MUTEX_BBC ) == Failure ) {
    gw_logf( LOG_SERIOUS, "Cannot release mutex" );
    return Failure;
  }

  return ret;
}

/* ------------------------------------------------------------------------- */
/*
 * Shutdown a server - sends a terminate message/kill signals
 */
static RetValue
shutdown_server( BootParams *bp, pid_t *pid )
{
  long       srvcls = 0;
  long       srvid  = 0;
  BBEntry   *bbe_p  = NULL; 
  GWHomedMsg hm     = {0};
  int    old_net_server = 0;

  GWMessage  reply  = {0};

  SHMLIST    *BB  = (SHMLIST*)   ((char*)bbc + bbc->BB_entry_offset );
  BBReqQueue *bbq = (BBReqQueue*)((char*)bbc + bbc->BB_queue_offset );

  srvcls = ((BootParams*)bp)->sp.servclass;
  srvid  = ((BootParams*)bp)->sp.servid;

  gw_logf( LOG_ALWAYS, "Shutdown server Class: %ld, Id: %ld...",srvcls,srvid );

  /*
   * Search BB for this server.
   */
  if( find_server_entry( local_dom->domainid, 
			 srvcls, srvid, &bbe_p, BB ) >= 0 ) {


    /* Are we a net server? 
     */
    if( strcmp( bbe_p->bb_sp.server, ONETSERVER ) == 0 ) old_net_server = 1;

    /* We don't want to respawn child... */
    bbe_p->bb_sc.respawn = CONTROLLED_SHUT; 

    *pid = bbe_p->bb_sc.pid; /* Return PID of server */

    /* Setup a false BBSvc block for the homed request
     */
    hm.servid  = srvid;
    hm.msgqid  = bbq[ bbe_p->req_queue_idx ].msgqid;
    hm.ident   = bbq[ bbe_p->req_queue_idx ].listen_ident;

    hm.msgtype = GWHM_HALT_SERVER;    /* What the Homed message is */
    
    if( old_net_server ) {
      /* KILL the network server with a signal
       */
      kill( *pid, SIGTERM );
      sleep(1); /* Safety net */
      
    } else {
      /* Send server a TERMINATE homed message 
       */
      if( gw_svc_call( (const char *)&hm, "", 1, &reply, GWP_HOMDMSG ) < 0 ) {
        resp_text = "Failed";
        return Failure;
      }
    }
    
    /* TODO: Should this be in the EXIT code of a server ?? */
    bbq[ bbe_p->req_queue_idx ].numdeps--; /* Decrement servers on queue */
    if( bbq[ bbe_p->req_queue_idx ].numdeps == 0 ) {
      /* Remove queue entry now there are no more servers attached to it */
      delete_queue_entry( bbe_p->req_queue_idx );
    }

    /* Sending of shutdown command was successful */
    resp_text = "Halted";
    return Success;
  }

  resp_text = "No server found";

  return Failure;
}

/* ------------------------------------------------------------------------- */
/*
 * Restarts a child that has died, or cleans up the BB if the child is not
 * supposed to be restarted (this cleanup includes releasing msgqueues when
 * they're finished with).
 */
static RetValue
restart_child( pid_t pid )
{
  RetValue ret = Success;
  int      class_set  = 0;
  int      msgqid     = -1;
  int      rmsgqid    = -1;
  int      cur_class  = -1;
  int      last_class = -1;
  long     bbqidx     = -1;
  uchar   *ptr        = NULL;
  BBEntry *bbe_p      = NULL;
  pid_t    spid       = 0; 

  SHMLIST    *BB  = (SHMLIST*)   ((char*)bbc + bbc->BB_entry_offset );
  SHMLIST    *BBS = (SHMLIST*)   ((char*)bbc + bbc->BB_svc_offset   );
  BBReqQueue *bbq = (BBReqQueue*)((char*)bbc + bbc->BB_queue_offset );

  /* Get Mutex */
  if( aquire_bbc_mutex( MUTEX_BBC ) == Failure ) return Failure;

  /* Find child and restart it...
   */
  if( shm_List_get_first( BB, &ptr ) == Success ) {
    do {
      bbe_p = (BBEntry*)ptr;

      cur_class = bbe_p->bb_sp.servclass;

      /* Found our server by matching pid's */
      if( bbe_p->bb_sc.pid == pid ) {

	/* Remove reply queue if there is one */
	rmsgqid = bbe_p->bb_sc.repmqid;
	if( rmsgqid >= 0 ) {
	  gw_log( LOG_INTDBG, "Remove Reply Queue", -1 );
	  remove_queue( rmsgqid );
	  bbe_p->bb_sc.repmqid = -1;
	}

	/* If the forced restarts (not due to restart_threshold being
	 * reached) get passed a limit then set this server to 
	 * Not_Respawnable (reset the bb_sc.respawn flag)
         *	
	 *	Child servers my exit because of a crash or because its' max
	 *	requests limit was reached
	 *	
         *   If the gracefull restart flag is not set, then the restart 
	 *   was due to a child dying and when restarts hit the 
	 *   'crash restart limit' we refuse to reboot it any more.
         */
        if( bbe_p->bb_sc.frstrt > MAX_SANE_RESTART && 
            bbe_p->bb_sc.gracerstrt_ind == 0 ) {

          bbe_p->bb_sc.respawn = DO_NOT_RESPAWN;
	  gw_logf( LOG_SYSWARN, 
	           "Maximum forced restart of server reached (%s : %ld : %d)",
                   bbe_p->bb_sp.server,bbe_p->bb_sp.servclass,
                   bbe_p->bb_sp.servid);
        }

	/* Are we allowed to respawn this server */
        if( bbe_p->bb_sc.respawn == ALLOW_RESPAWN ) {

	  gw_logf( LOG_ALWAYS, "%s Server '%s' (Class %ld, Id %d)",
	  	   bbe_p->bb_sc.gracerstrt_ind ? "Cycling" : "Restart",
		   bbe_p->bb_sp.server,
		   bbe_p->bb_sp.servclass,
		   bbe_p->bb_sp.servid );
	
	  if( chdir( bbc->appdir ) < 0 ) {
	    gw_logf( LOG_SERIOUS, "Could not access EXEDIR %s", bbc->appdir );
	    return Failure;
	  }
	  

          /* Release The Mutex Before launching server!!!
	   */
          if( release_bbc_mutex( MUTEX_BBC ) == Failure ) return Failure;

	  ret = exec_server( bbe_p->bb_sp.server, 
			     local_dom->domain, 
			     bbq[ bbe_p->req_queue_idx ].msgqid,
			     bbe_p->bb_sp.servid, 
			     bbe_p->bb_sp.servclass, 
			     &spid );

          /* Get The Mutex back again 
	   * (This drop and pickup phaf could be avoided with 
           * some cleaner code)
	   */
          if( aquire_bbc_mutex( MUTEX_BBC ) == Failure ) return Failure;

	  if( ret == Success ) {
	    bbe_p->bb_sc.pid    = spid; /* Reset PID                 */

            if( bbe_p->bb_sc.gracerstrt_ind == 0 ) {
	      bbe_p->bb_sc.frstrt++;      /* Increment Forced Restarts */
            }
            strcpy( bbe_p->bb_sc.state, "Ready" );
	  }
	} else {

	  /* This code is duplicated in clean_bb_for_server()
	   * Making a call to that function from here instead of duplicating
	   * is not a good idea because it does another scan of the BB!!!!!!
	   *
	   * PLUS we can get in a real mess with mutex's
	   *
	   * We've got the info we want right now, so we may as well do the
	   * cleanup on our own.
	   */
	  if( bbe_p->bb_sc.respawn != CONTROLLED_SHUT ) {
	    /* Restart of server is disabled.  So clean up BB and BBS */

	    gw_logf( LOG_INTDBG, 
		     "Restart of server '%s' (Class %ld, Id %d) disabled",
		     bbe_p->bb_sp.server,
		     bbe_p->bb_sp.servclass,
		     bbe_p->bb_sp.servid );
	    
        bbqidx = bbe_p->req_queue_idx;
	    msgqid = bbq[ bbqidx ].msgqid;

	    if( last_class >= 0 && last_class == cur_class ) class_set = 1;

	    /* Remove services for this server
	     */
	    __gw_remove_services( bbc->domainid, 
                                  bbe_p->bb_sp.servclass, 
                                  bbe_p->bb_sp.servid, 
                                  BBS );
	    
	    /* Remove entry from BB
	     */
	    gw_logf( LOG_INTDBG, 
		     "Removing BB entry for server pid %ld", 
                     (long)bbe_p->bb_sc.pid );
	    
	    shm_List_delete_node( BB );

	    /* We need to do a quick check to see if the NEW current node is
	     * of the same class or, if not, the next node.
	     * This is because of the delete_node behaviour differing when we 
	     * are deleting from the head of a list or the middle/end.  
	     * If it is the same class, then we've still got a class set.
	     */
	    bbe_p = (BBEntry*)shm_List_get_current_data;
	    
	    if( cur_class == bbe_p->bb_sp.servclass ) class_set = 1;
	    else if( shm_List_get_next( BB, (uchar**)&bbe_p ) == Success &&
		     cur_class == bbe_p->bb_sp.servclass ) class_set = 1;

	    /* Can we remove the message queue? */
	    if( ! class_set && msgqid >= 0) {
	      gw_logf( LOG_INTDBG, "Removing message queue %ld", msgqid );
	      remove_queue( msgqid );
          delete_queue_entry( bbqidx );
	    }
	  }
	}
	bbe_p->bb_sc.gracerstrt_ind = 0;
	break;
      }

      last_class = cur_class;

    } while( shm_List_get_next( BB, &ptr ) == Success ) ;
  }

  /* Release Mutex */
  if( release_bbc_mutex( MUTEX_BBC ) == Failure ) return Failure;

  return ret;
}


/* ------------------------------------------------------------------------- */
/*
 * Given the PID of a server, this function removes any services registered
 * for it along with it's BB entry.
 *
 * Clean only occurs if bbe->bb_sc.respawn is CONTROLLED_SHUT
 *
 */
static RetValue
clean_bb_for_server( pid_t pid )
{
  RetValue ret = Success;
  int      class_set  = 0;
  int      msgqid     = -1;
  int      rmsgqid    = -1;
  int      cur_class  = -1;
  int      last_class = -1;
  long     bbqidx     = -1;
  uchar   *ptr        = NULL;
  BBEntry *bbe_p      = NULL;

  SHMLIST    *BB  = (SHMLIST*)   ((char*)bbc + bbc->BB_entry_offset );
  SHMLIST    *BBS = (SHMLIST*)   ((char*)bbc + bbc->BB_svc_offset   );
  BBReqQueue *bbq = (BBReqQueue*)((char*)bbc + bbc->BB_queue_offset );

  /* Get Mutex */
  if( aquire_bbc_mutex( MUTEX_BBC ) == Failure ) return Failure;

  /* Find child and restart it...
   */
  if( shm_List_get_first( BB, &ptr ) == Success ) {
    do {
      bbe_p = (BBEntry*)ptr;

      /* As entries are ordered on CLASS we can keep a simple count on
       * whether we have multiple entries for this class.
       * We're not particularly interested in the actual number, just that
       * we have more than one.  We use this to decide whether to remove the
       * message queue or not.
       */
      cur_class = bbe_p->bb_sp.servclass;

      /* if we find our server (by pid) and we're to shut it down... .*/
      if( bbe_p->bb_sc.pid == pid ) {
	if( bbe_p->bb_sc.respawn == CONTROLLED_SHUT ) {

          bbqidx  = bbe_p->req_queue_idx;
	  msgqid  = bbq[ bbqidx ].msgqid;
	  rmsgqid = bbe_p->bb_sc.repmqid;

	  if( last_class >= 0 && last_class == cur_class ) class_set = 1;

	  /* Remove services for this server (in the local domain)
	   */
	  __gw_remove_services( bbc->domainid, cur_class, 
                                bbe_p->bb_sp.servid, BBS);
	  
	  /* Remove entry from BB
	   */
	  gw_logf( LOG_INTDBG,  "Removing BB entry for server pid %ld", 
		   (long)bbe_p->bb_sc.pid );
	  
	  shm_List_delete_node( BB );

	  /* We need to do a quick check to see if the NEW current node is
	   * of the same class or, if not, the next node.
	   * This is because of the delete_node behaviour differing when we 
	   * are deleting from the head of a list or the middle/end.  
	   * If it is the same class, then we've still got a class set.
	   */
	  bbe_p = (BBEntry*)shm_List_get_current_data;
	  
	  if( cur_class == bbe_p->bb_sp.servclass ) class_set = 1;
	  else if( shm_List_get_next( BB, (uchar**)&bbe_p ) == Success &&
		   cur_class == bbe_p->bb_sp.servclass ) class_set = 1;

	  /* Can we remove the (reply) message queue? (do we have one?) */
	  if( rmsgqid >= 0 ) remove_queue( rmsgqid );
	  
	  if( ! class_set && msgqid >= 0) {
	    /* No more servers in class, We can ditch the request queue */
	    gw_logf( LOG_INTDBG, "Removing message queue %ld", msgqid );
	    remove_queue( msgqid );
	    delete_queue_entry( bbqidx ); 
	  }
	}
	break;
      }

      last_class = cur_class;

    } while( shm_List_get_next( BB, &ptr ) == Success ) ;
  }

  /* Release Mutex */
  if( release_bbc_mutex( MUTEX_BBC ) == Failure ) return Failure;

  return ret;
}

/* ------------------------------------------------------------------------- */

static long
create_queue_entry( long msgqid )
{
  long idx;

  BBReqQueue *bbq = (BBReqQueue*)((char*)bbc + bbc->BB_queue_offset );

#ifdef DEBUG
  /* Count number of entries in queue array */
  int count = 0;
  for( idx = 0; idx < bbc->queuemax; idx++ ) {
    if( bbq[ idx ].msgqid >= 0 ) count++; 
  }
  gw_logf( LOG_ALWAYS, "Used slots in BBReqQueue: %d", count );
#endif

  /* Find first unused BBReqQueue entry (this is an array!)
   * - It's an array becase we want referencing to be FAST
   *   We don't generally search the array, only when looking for an
   *   empty slot for a NEW entry when booting.
   */
  for( idx = 0; idx < bbc->queuemax; idx++ ) {
    if( bbq[ idx ].msgqid < 0 ) break;
  }

  if( idx == bbc->queuemax ) {
    gw_logf( LOG_SERIOUS, "No room for new Request Queue" );
    return -1;
  }


  /* Set initial values
   */
  bbq[ idx ].msgqid       = msgqid;
  bbq[ idx ].listen_ident = MAX_REQ_LISTEN_ID;
  bbq[ idx ].numonqueue   = 0;
  bbq[ idx ].aveQtime     = 0;
  bbq[ idx ].numdeps      = 0;

  return idx;
}

/* ------------------------------------------------------------------------- */

static void
delete_queue_entry( long qidx )
{
  BBReqQueue *bbq = (BBReqQueue*)((char*)bbc + bbc->BB_queue_offset );

  memset( (void*)&bbq[ qidx ], 0, sizeof(BBReqQueue));
  bbq[ qidx ].msgqid = -1;
}

/* ------------------------------------------------------------------------- */
/*
 * Handles signals from child, in particular SIGCHLD or CHILD_OKAY_SIG
 */
static void
child_signal( int sig )
{
  pid_t cpid = 0;

  switch( sig ) {
    
  case SIGCHLD:
    /* Pick up as many child signals as we can....
	 */
    while( (cpid = waitpid( -1, NULL, WNOHANG)) > 0 ) {

      gw_log_init( _ServId, "Guardian" ); 
      gw_logf( LOG_ALWAYS, "Server (%ld) terminated", (long)cpid );
    
      /* An already running server has died... restart it if we can. */
      if( cpid > 0 && cpid != booting_child_pid ) restart_child( cpid );
    }
    break;

  case CHILD_OKAY_SIG:
    siglongjmp( jbuf, CHILD_READY ); 
    break;

  default: /* Report all other signals via child_sig */
    child_sig = sig;
    break;
  }
}

/* ------------------------------------------------------------------------- */

static RetValue
exec_server( const char *server, const char *domain, 
	     int msgqid, int servid, int sclass, pid_t *spid )
{
  char *args[20]  = {0};
  int   i         =  0;
  sigset_t smask, oldsmask;
  int old_net_server  = 0;
  int sys_server  = 0;

  child_sig   = 0;
  child_error = 0;

  if( strcmp( server, ONETSERVER ) == 0 ) old_net_server = 1;
  if( strcmp( server, SYSSERVER1 ) == 0 ) sys_server = 1;
  if( strcmp( server, SYSSERVER2 ) == 0 ) sys_server = 1;

  /* Block SIGCHLD, Unblock CHILD_OKAY_SIG signals whilst in this function...
   */
  sigemptyset( &smask );
  sigaddset  ( &smask, SIGCHLD );
  sigprocmask( SIG_BLOCK, &smask, &oldsmask );

  sigemptyset( &smask );
  sigaddset  ( &smask, CHILD_OKAY_SIG );
  sigprocmask( SIG_UNBLOCK, &smask, NULL );

  switch( sigsetjmp( jbuf, 1 ) ) {
    
  case SIGJMP_OKAY: /* the 0 return that sigsetjmp returns with directly... */
    /* Just get on with the fork below... */
    break;

  case CHILD_READY:
    *spid = booting_child_pid;
    booting_child_pid = -1;

    /* Restore Signals
     */
    sigprocmask( SIG_SETMASK, &oldsmask, NULL );

    gw_logf( LOG_SYSINFO, "Server booted (%ld)", (long)*spid );

    return Success;   /* Exit Point 1 */
    break;

  case CHILD_FAILURE: /* Genuine Failure of server to start successfully */
  default:            /* Unknown return from sigsetjmp... */

    /* Restore Signals
     */
    sigprocmask( SIG_SETMASK, &oldsmask, NULL );
    
    return Failure;   /* Exit Point 2 */
    break;
  }

  /* Now fork/exec the server and wait for it to start....
   */
  switch( (booting_child_pid = vfork()) ) {
  case 0:  /* child */
    {
      char servid_s[64] = {0};
      char sclass_s[64] = {0};
      char msgqid_s[64] = {0};
      char ipckey_s[64] = {0};
      
      sprintf( servid_s, "%d",  servid );
      sprintf( sclass_s, "%d",  sclass );
      sprintf( msgqid_s, "%d",  msgqid );
      sprintf( ipckey_s, "%ld", _IpcKey );

      /* 
       * Initialise Arguments 
       */
      i = 0;
#if VALGRIND
if( strcmp( server, "DGSvcMgr" ) == 0 ) {
args[i++] = "/usr/local/bin/valgrind";
args[i++] = "--leak-check=yes";
args[i++] = "--logfile=/tmp/vlog";
args[i++] = "--show-reachable=yes";
args[i++] = "--verbose";
args[i++] = "--verbose";
args[i++] = "/usr/local/libexec/dgee/DGSvcMgr";
} else {
      args[i++] = (char *)server;
}
#else
      args[i++] = (char *)server;
#endif
      args[i++] = (char*)domain;
      args[i++] = servid_s;
      args[i++] = sclass_s;
      args[i++] = msgqid >= 0 ? msgqid_s : "0";
      args[i++] = ipckey_s;
      args[i++] = NULL;
#if 0
      {
        int d = 0;
        while( d < i ) {
          gw_logf( LOG_INTDBG, "Param %d: '%s'", d, args[d] );
          d++;
        }
      }
#endif
      /* Unblock all signals for child 
       */
      sigfillset( &smask );
      sigprocmask( SIG_UNBLOCK, &smask, NULL );

      /* Close the open logfile as the 'server' will open it itself anyway */
      gw_closelog();

      /* If we're a system server then cd to the installation bin dir.
       * This is a security measure as we may be running as ruid root and
       * thus don't want to go searching the path for our server exe!
       * We know where our system server are at install time, so we use that
       * directory here.
       */
      if( (old_net_server||sys_server) && chdir(GWSysBin) < 0 ) {
        gw_logf( LOG_SERIOUS, "Could not change directory to %s", GWSysBin );
      } 
      else execv(args[0], args);

      gw_logf( LOG_SYSERR, "Failed to launch '%s'", args[0] );

      /* 
       * The server is probably not a binary then.  Someone's been stupid.
       */

      /* We must use _exit() here and not exit() because we DO NOT want to call
       * the exit functions registered with atexit() or on_exit()
       */
      _exit( COULD_NOT_EXEC ); /* We've got to terminate this child process. */
    }
    break;
    
  case -1: /* Failure */
    
    siglongjmp( jbuf, CHILD_FAILURE ); /* go to setsigjmp */
    break;
    
  default: /* Parent */
    /* Wait for a signal from child.  Will either be SIGCHLD or CHILD_SIG_OKAY
     * - If a child DOES fail, as we're blocking SIGCHLD we won't get this
     *   signal for a while.... thus this sleep will complete - a small price
     *   to pay for failing servers... which'll be the rarer case.
     */
    sleep( _start_timer );

    /* We didn't get a startup signal within the timeout period....
     * Kill child off.. we're not going to get this signal until we unblock
     * SIGCHLD.
     */
    kill( booting_child_pid, SIGTERM );


    siglongjmp( jbuf, CHILD_FAILURE ); /* go to setsigjmp */

  }
  
  /* exit points are in the setjmp switch above... */
  
}

/* ------------------------------------------------------------------------- */
/* end */

