/*
 * Goldwater Distributed Web Architecture.
 *
 * Log Handling
 *
 * Copyright 1996-2002 netFluid Technology Ltd
 *
 * Goldwater is written and maintained by netFluid Technology Ltd.
 * http://www.nfluid.co.uk
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 *
 * $Revision: 1.4 $  $Date: 2003/07/19 11:25:24 $
 * 
 * --------------------------------------------------------------------------
 * $Log: gwlog.c,v $
 * Revision 1.4  2003/07/19 11:25:24  csmith
 * Report to stderr when logfiles cannot be opened
 *
 * Revision 1.3  2003/06/12 11:56:09  csmith
 * C# wrappers
 *
 * Revision 1.2  2003/03/29 12:21:11  csmith
 * cvs merge to 0.3.4
 *
 * Revision 1.23  2003/01/05 22:03:47  chris
 * Open logfile on init
 *
 * Revision 1.22  2003/01/04 19:34:11  chris
 * gw_init() is now quite intelligent with respect to reconnecting.
 *
 * Revision 1.21  2003/01/03 00:09:11  chris
 * Use internal umask when creating files, shm and queues
 *
 * Revision 1.20  2003/01/02 18:49:56  chris
 * Added User and Group
 *
 * Revision 1.19  2002/11/11 20:26:34  chris
 * implemented gw_svc_forward.
 *
 * Revision 1.18  2002/11/10 12:23:41  chris
 * Cleanup and low level API changes for bindings such as C#
 *
 * Revision 1.17  2002/10/20 13:38:34  chris
 * Enabled setting of loglevel from administrator (removed _LogLevel and
 * replaced it with *_LogLevelPtr through out code).
 *
 * Revision 1.16  2002/09/01 10:59:31  chris
 * Addition of namespaceing to service name.
 * Service may now be of the form this.is.a.namespace/and/this/is/the/service
 * Wildcards are NOT allowed in RPC type message destinations.
 *
 * Revision 1.15  2002/06/16 11:40:35  chris
 * BB Client is now the Domain Bridge message tracking list.
 * New error codes and messages.
 *
 * Revision 1.14  2002/06/15 18:54:39  chris
 * Domain Bridge additions
 *
 * Revision 1.13  2002/06/10 17:34:47  chris
 * All phlib includes have phlib/ prefix and gwlib includes have nothing.
 *
 * Revision 1.12  2002/05/30 19:37:34  chris
 * Create entries in BBClient table for each new client attaching to
 * application.
 *
 * Revision 1.11  2002/05/12 22:59:02  chris
 * Added new error GWE_UNSTABLE - returned when GW detected to be in an
 * unstable state.
 *
 * Revision 1.10  2002/05/12 14:03:13  chris
 * GPL licence headers etc
 *
 * Revision 1.9  2002/02/28 00:55:16  chris
 * Various bug fixes.
 *
 * Revision 1.8  2002/02/24 18:05:38  chris
 * Moved APPDIR to BBControl block instead of duplicating it for each and
 * every server.
 * Return GWE_APPERR and set gwusrerr when Services return failure.
 * Use new regstrings that describe the new XML based configuration.
 * Added new functions to support individual addition of service entries.
 * New Version 1.14
 *
 * Revision 1.7  2002/01/29 22:43:40  chris
 * Cleaned up warnings.
 * Added gwp_ prefix to shutdown_server.
 * Added regdata.[ch] config entry to build.
 *
 * Revision 1.6  2002/01/03 20:27:31  chris
 * Introduction of preliminary Domain Control handling.
 *
 * Revision 1.5  2001/12/26 22:57:43  chris
 * Changed all references to Imagepage Computing to netFluid Technology.
 *
 * Revision 1.4  2001/03/30 22:16:01  chris
 * Added GWE_BADHDR error and desc string.
 *
 * Revision 1.3  2001/03/30 19:22:05  chris
 * Now writes out date and time.
 *
 * Revision 1.2  2001/03/22 01:14:02  chris
 * Added new function gw_log and ripped out the log-preamble into an
 * inline-ing function.  gw_log used writev, avoiding the need for
 * multiple write calls or lots of memcpy's.
 *
 * Revision 1.1  2001/03/19 16:05:36  chris
 * Initial revision
 *
 * --------------------------------------------------------------------------
 */
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/uio.h>

#include <phlib/xl_date.h>

#include <gwreq.h>
#include <gwgbprm.h>

#define NEED_ERROR_STRINGS 1
#  include <gwlog.h>
#undef  NEED_ERROR_STRINGS

static int date_init = 0;

static long _logpid                     = -1;
static char _logservid[11]              = {0};
static char _logservice[SVC_NM_MAX + 1] = {0};

static unsigned char *_date_str = "00/00/00";
static unsigned char *_time_str = "00:00:00";

static int logfd = -1;

static char logbuf[16384];
 
static int gw_open_log( const char * );
static __inline__ int log_preamble( long level );

/* ------------------------------------------------------------------------- */

static __inline__
int log_preamble( long level )
{
  int l = 0;

  xl_get_time(1);  /* Refresh both date and time strings... */

  /* Log format:
   *
   * date time:pid:servid:service:LogLevel:text
   */
  l = sprintf( logbuf, "%s %s %5.5ld:%s:%s:",
	       _date_str,
	       _time_str,
	       _logpid,
	       _logservid,
	       _logservice );

  if( (level & LOG_ALWAYS) == LOG_ALWAYS ) { 
    memcpy( logbuf + l, "ADM", 4 ); 
    l+=4; 
    level = 0;
  }

  if( level & LOG_ERROR   ) { memcpy( logbuf + l, "ERR|", 4 ); l+=4; }
  if( level & LOG_DATA    ) { memcpy( logbuf + l, "DTA|", 4 ); l+=4; }
  if( level & LOG_FLOW    ) { memcpy( logbuf + l, "FLO|", 4 ); l+=4; }
  if( level & LOG_DEBUG   ) { memcpy( logbuf + l, "DBG|", 4 ); l+=4; }
  if( level & LOG_INFO    ) { memcpy( logbuf + l, "INF|", 4 ); l+=4; }
  if( level & LOG_AUDIT   ) { memcpy( logbuf + l, "AUD|", 4 ); l+=4; }

  if( level & LOG_INTDBG  ) { memcpy( logbuf + l, "IDB|", 4 ); l+=4; }

  if( level & LOG_SYSERR  ) { memcpy( logbuf + l, "ERR|", 4 ); l+=4; }
  if( level & LOG_SYSINFO ) { memcpy( logbuf + l, "INF|", 4 ); l+=4; }
  if( level & LOG_SYSWARN ) { memcpy( logbuf + l, "WRN|", 4 ); l+=4; }
  if( level & LOG_SERIOUS ) { memcpy( logbuf + l, "SER|", 4 ); l+=4; }
  
  logbuf[l-1] = ' ';
  logbuf[l++] = '>';
  logbuf[l++] = ' ';

  return l;
}

/* ------------------------------------------------------------------------- */
/*
 * Writes a 'len' length string to the log (with the ususal preamble).
 * if len < 0, then the length of str is calculated and used instead.
 */
void
gw_log( long level, char *str, int len )
{
  int l = 0;
  static struct iovec lv[3];
  if( !_LogLevelPtr || !((*_LogLevelPtr | LOG_ALLSYS) & (long)level) ) return;
  if( logfd < 0 && gw_open_log( _LogName ) < 0 ) return;
  
  l = log_preamble( level );
  
  lv[0].iov_base = logbuf;
  lv[0].iov_len  = l;

  lv[1].iov_base = str;
  lv[1].iov_len  = (len >= 0) ? len : strlen(str);

  lv[2].iov_base = "\n";
  lv[2].iov_len  = 1;

  writev( logfd, lv, 3 );
}

/* ------------------------------------------------------------------------- */

void
gw_logf( long level, char *fmt, ... )
{
  int l = 0;
  int m = 0;
  va_list ap;

  if( !_LogLevelPtr || !((*_LogLevelPtr | LOG_ALLSYS) & (long)level) ) return;
  if( logfd < 0 && gw_open_log( _LogName ) < 0 ) return;

  l = log_preamble( level );

  va_start( ap, fmt );
  m = vsprintf( logbuf+l, fmt, ap );
  va_end  ( ap );
  
  logbuf[l+m] = '\n';

  write( logfd, logbuf, l + m + 1 );
}

/* ------------------------------------------------------------------------- */

void
gw_closelog()
{
  if( logfd >= 0 ) close( logfd );
}

/* ------------------------------------------------------------------------- */
/*
 * Reset all pre-init'ed log settings
 */
void
gw_log_reset()
{
  date_init = 0;
  _logservice[0] = '\0';
  _logservid[0]  = '\0';
  _logpid        = -1;
}

/* ------------------------------------------------------------------------- */

void
gw_log_init( long servid, const char *service )
{
  if( date_init == 0 ) {
    xl_init_date();
    _date_str = xl_get_date(0);
    _time_str = xl_get_time(0);

    date_init = 1;
  }

  _logservice[0] = '\0';
  if( service != NULL ) strcpy ( _logservice, service);
  
  if( _logservid[0] == '\0' && servid > 0 ) 
    sprintf( _logservid, "%ld", servid );
  
  if( _logpid == -1 ) _logpid = getpid();

  if( logfd < 0 && gw_open_log( _LogName ) < 0 ) return;
}

/* ------------------------------------------------------------------------- */

static int
gw_open_log( const char *lfile ) 
{
  mode_t om;

  if( lfile == NULL || lfile[0] == '\0' ) return -1;

  om = umask( _Mask );

  logfd = open( lfile, O_WRONLY | O_CREAT | O_APPEND | O_NONBLOCK, 0666&~_Mask);

  umask( om );

  if( logfd < 0 ) {
	  fprintf( stderr, "Cannot open logfile %s for writing\n", lfile);
	  return -1;
  }
  
  gw_logf( LOG_INTDBG, "Opened logfile with fd %d", logfd );
  
  return 0;
}

/* ------------------------------------------------------------------------- */

const char *
gw_errorstr( long error )
{
  const char *str = "<no translation>";

  switch( error & GWE_ERRMASK ) {
  case GWE_NOFUNC:  str = ES_NOFUNC ; break;
  case GWE_TIMEOUT: str = ES_TIMEOUT; break;
  case GWE_GENFAIL: str = ES_GENFAIL; break;
  case GWE_NOTCNCT: str = ES_NOTCNCT; break;
  case GWE_READERR: str = ES_READERR; break;
  case GWE_GWUAVAL: str = ES_GWUAVAL; break;
  case GWE_DISCNCT: str = ES_DISCNCT; break;
  case GWE_BADDATA: str = ES_BADDATA; break;
  case GWE_BADCNFG: str = ES_BADCNFG; break;
  case GWE_CFGNFND: str = ES_CFGNFND; break;
  case GWE_SYSFAIL: str = ES_SYSFAIL; break;
  case GWE_NOSVC  : str = ES_NOSVC  ; break;
  case GWE_SBLOCKD: str = ES_SBLOCKD; break;
  case GWE_REPSENT: str = ES_REPSENT; break;
  case GWE_REPNSNT: str = ES_REPNSNT; break;
  case GWE_BADHDR : str = ES_BADHDR ; break;
  case GWE_SPLDIR : str = ES_SPLDIR ; break;
  case GWE_APPERR : str = ES_APPERR ; break;
  case GWE_CFGRERR: str = ES_CFGRERR; break;
  case GWE_BCFGARC: str = ES_BCFGARC; break;
  case GWE_BADCFG : str = ES_BADCFG ; break;
  case GWE_NOTUP  : str = ES_NOTUP  ; break;
  case GWE_STILLUP: str = ES_STILLUP; break;
  case GWE_CLTFAIL: str = ES_CLTFAIL; break;
  case GWE_UNSTABLE:str = ES_UNSTABLE; break;
  case GWE_BADDMVER:str = ES_BADDMVER; break;
  case GWE_BADIPCKY:str = ES_BADIPCKY; break;
  case GWE_BADLOGF :str = ES_BADLOGF ; break;
  case GWE_BADDADDR:str = ES_BADDADDR; break;
  case GWE_BADNSPC :str = ES_BADNSPC ; break;
  case GWE_BADUSER :str = ES_BADUSER ; break;
  case GWE_BADGROUP:str = ES_BADGROUP; break;
  case GWE_SETUSER :str = ES_SETUSER ; break;
  case GWE_SETGROUP:str = ES_SETGROUP; break;
  default: break;
  }
  return str;
}

/* ------------------------------------------------------------------------- */

const char *
gw_last_errorstr()
{
  return gw_errorstr( gwerror );
}

/* ------------------------------------------------------------------------- */

int
__gw_cs_get_gwerror()
{
  return gwerror;
}

/* ------------------------------------------------------------------------- */

int
__gw_cs_get_gwusrerr()
{
  return gwusrerr;
}

/* ------------------------------------------------------------------------- */
/* end */
