/*
 * GWPortal Comms
 *
 * Copyright 1996-2002 netFluid Technology Ltd
 *
 * Goldwater is written and maintained by netFluid Technology Ltd.
 * http://www.nfluid.co.uk
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 *
 * $Revision: 1.2 $
 *
 * ----------------------------------------------------------------------- 
 * $Log: gwportal.c,v $
 * Revision 1.2  2003/03/29 12:21:11  csmith
 * cvs merge to 0.3.4
 *
 * Revision 1.9  2002/11/10 12:23:41  chris
 * Cleanup and low level API changes for bindings such as C#
 *
 * Revision 1.8  2002/06/15 19:32:05  chris
 * Message ID values are now 128 bit OSF DCE compatible uuid's.
 *
 * Revision 1.7  2002/06/10 17:34:47  chris
 * All phlib includes have phlib/ prefix and gwlib includes have nothing.
 *
 * Revision 1.6  2002/05/12 14:03:13  chris
 * GPL licence headers etc
 *
 * Revision 1.5  2002/03/10 19:35:33  chris
 * Handles proper homed messages (initially for shutdown of servers).
 *
 * Revision 1.4  2002/02/24 18:05:38  chris
 * Moved APPDIR to BBControl block instead of duplicating it for each and
 * every server.
 * Return GWE_APPERR and set gwusrerr when Services return failure.
 * Use new regstrings that describe the new XML based configuration.
 * Added new functions to support individual addition of service entries.
 * New Version 1.14
 *
 * Revision 1.3  2001/12/26 22:57:43  chris
 * Changed all references to Imagepage Computing to netFluid Technology.
 *
 * Revision 1.2  2001/04/01 22:49:25  chris
 * Changed text of some failure logs.
 *
 * Revision 1.1  2001/03/19 16:06:23  chris
 * Initial revision
 *
 * ----------------------------------------------------------------------- 
 *
 */
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#define GW_INTERNAL_FLAGS 1

#include <phlib/xl_types.h>
#include <phlib/xl_string.h>
#include <phlib/xl_socket.h>
#include <phlib/xl_memory.h>

#include <gwportal.h>
#include <gwreq.h>
#include <gwcmdstr.h>
#include <gwgbprm.h>
#include <gwlog.h>

static int rd_socket = (-1);
int gwerror  = 0;
int gwusrerr = 0;

/* ----------------------------------------------------------------------- */
/*
 * Return Values: 0 = connected
 *               -1 = Failure
 *                    rderror gives reason for failure
 *                    GWPNOSERVER
 *                    GWPREADERR
 */
int
gwp_portal_connect( const char *host, int port )
{
  int    ret     = MINUSONE;
  int    l       = 0;
  char   buf[128] ={0}; 

  gwerror = 0;
  
  if( rd_socket >= 0 ) return 0; /* We're connected already */

  gwerror = GWE_GWUAVAL;

  if( (rd_socket = xl_get_connection( host, port ) ) >= 0 ) {

    l = read( rd_socket, buf, sizeof(buf) );
    
    if( strncmp( buf, GREET_STR, sizeof( GREET_STR ) -1 ) != 0 ) {
      close( rd_socket );
      gw_logf( LOG_SERIOUS, 
	       "Did not receive expected greeting "
	       "when connecting to scheduler" );

      rd_socket = (-1);
      gwerror = GWE_READERR;
    }

    ret = 0;
    gwerror = 0;
  }  else {
    _gw_initialised--;
    gw_logf( LOG_SYSERR, "Cannot connect to Goldwater" );
  }

  return ret;
}


/* ----------------------------------------------------------------------- */

void
gwp_portal_disconnect()
{
  gwerror = 0;
  if( rd_socket >= 0 ) {
    /* terminate_connection( rd_socket ); */
    close( rd_socket );
    rd_socket = (-1);
    gw_logf( LOG_INTDBG, "Disconnection from Scheduler (portal)" );
  }
}

/* ----------------------------------------------------------------------- */

int
gwp_setup_header( const GWMsgCtrl *ctrl, char *buf, long len )
{
  long  slen = 0;
  char *cmd  = "";

  if( ctrl->flags & GWP_REPLY ) {
    cmd = ".REP";  /* Passing reply for client to portal */
  } else {
    cmd = ".REQ";  /* Pass request to a service          */
    slen = strlen( ctrl->service ) + 1; /* +1 for \0 */
  }
  strncpy( buf, cmd, 4 );
  
#if 0 /* THIS IS BROKEN BY NEW USE OF UUID For MSGID */
  store_int32_LittleEndian( buf + 4,  ctrl->msgid      ); /* Message ident   */
#else
  store_int32_LittleEndian( buf + 4,  getpid()   ); /* Message ident   */
#endif
  store_int32_LittleEndian( buf + 8,  ctrl->clid       ); /* reply for clid  */
  store_int32_LittleEndian( buf + 12, ctrl->flags      );
  store_int32_LittleEndian( buf + 16, ctrl->total_size ); /* Total msg size  */
  store_int32_LittleEndian( buf + 20, len              ); /* of folwng data  */
  store_int32_LittleEndian( buf + 24, slen             ); /* of service name */

  /* Give Request Service */
  if(slen) memcpy( buf + 28, ctrl->service, slen ); 

  buf[28 + slen + 1] = '\r';
  buf[28 + slen + 2] = '\n';

  return slen + 28 + 2;
}

/* ----------------------------------------------------------------------- */

long
gwp_process_reply_header( GWMsgCtrl *ctrl, const char *buf, long len )
{
  long slen = 0;
  long dlen = 0;

  if( len < 28 ) return -1;
  if( strncmp( buf, ".REP", 4 ) != 0 ) return -1;

  /* BROKEN BY USE OF UUID AS MSGID */
  /*ctrl->msgid      =  read_int32_LittleEndian( (char*)buf + 4  ); */

  uuid_clear( ctrl->msgid );

  ctrl->clid       = read_int32_LittleEndian( (char*)buf + 8  );
  ctrl->flags      = read_int32_LittleEndian( (char*)buf + 12 );
  ctrl->total_size = read_int32_LittleEndian( (char*)buf + 16 );
  dlen             = read_int32_LittleEndian( (char*)buf + 20 );
  slen             = read_int32_LittleEndian( (char*)buf + 24 );

  /* There should be no service name in the header.... 
   * If there is we error.... */
  if( slen + 28 > len ) {
    gw_logf( LOG_SERIOUS, "Bad service length in request" );
    return -1; /* Bad service length */
  }

  return slen + 28;
}

/* ----------------------------------------------------------------------- */

int
gwp_send_to_portal( const GWMsgCtrl *ctrl, const char *data, long len )
{
  int   l         = 0;
  long  r         = 0;
  long  slen      = 0;
  const char *s   = NULL;
  char  rbuf[128] = {0};
  char  cbuf[128] = {0};

  gw_logf( LOG_INTDBG, "Send to portal (%ld bytes)", len );

  /*
   * Write command header to Phoenix 
   */

  if( rd_socket < 0 ) {
    /* We're not connected! */

    gw_logf( LOG_SYSERR, "No connection to Goldwater Scheduler" );
    gwerror = GWE_NOTCNCT;
    return MINUSONE;
  }

  slen = gwp_setup_header( ctrl, cbuf, len );

  /*
   * Write command header to Phoenix 
   */
  l = write( rd_socket, cbuf, slen );

  if( l <= 0 ) {
    close( rd_socket );
    rd_socket = MINUSONE;
    gw_logf( LOG_SERIOUS, "Unexpected disconnection from Goldwater Scheduler");
    gwerror = GWE_DISCNCT;
    return MINUSONE;
  }

  /*
   * Now read response
   */
  l = read( rd_socket, rbuf, 6 );

  if( l < 6 ) {
    /* Connection probably dropped */
    close( rd_socket );
    rd_socket = MINUSONE;
    gw_logf( LOG_SERIOUS, "Unexpected disconnection from Goldwater Scheduler");
    gwerror = GWE_DISCNCT;
    return MINUSONE;
  }

  /*
   * Make sure gwportal responded asking us for the data segment...
   */
  if( strncmp( DATA_RSP, rbuf, sizeof(DATA_RSP) - 1 ) != 0 ) {

    /* Okay convert the 4 character reason string (numeric) into a real
     * number and return this as the error code
     */
    rbuf[4] = '\0';
    gwerror = atoi( rbuf );
    gw_logf( LOG_SERIOUS, "Failure received from Goldwater Scheduler: %ld",
	     gwerror );
    return MINUSONE;
  }


  /* Okay, send the data segment
   */
  s = data;
  r = len;
  do {
    l = write( rd_socket, s, r );
    
    if( l <= 0 ) {
      close( rd_socket );
      rd_socket = MINUSONE;
      gwerror = GWE_DISCNCT;
      gw_logf( LOG_SERIOUS, 
	       "Unexpected disconnection from Goldwater Scheduler");
      return MINUSONE;
    }

    s += l;
    r -= l;
  } while( r > 0 );


  /*
   * Now we wait for a DATA_ACK_RSP....
   */
  l = read( rd_socket, rbuf, 6 );

  if( l < 6 ) {
    /* Connection probably dropped */
    close( rd_socket );
    rd_socket = MINUSONE;
    gwerror = GWE_DISCNCT;
    gw_logf( LOG_SERIOUS, 
	     "Unexpected disconnection from Goldwater Scheduler");
    return MINUSONE;
  }

  /*
   * Make sure gwportal responded with an ack
   */
  if( strncmp( DATA_ACK_RSP, rbuf, sizeof(DATA_ACK_RSP) - 1 ) != 0 ) {
    rbuf[4] = '\0';
    gwerror = atoi( rbuf );
    gw_logf( LOG_SERIOUS, "Failure received from Goldwater Scheduler: %ld",
	     gwerror );
    return MINUSONE;
  }
  
  gwerror = 0;

  return 0;
}


/* ----------------------------------------------------------------------- */

char *
gwp_get_reply_from_portal( GWMsgCtrl *ctrl )
{
  char header[32];
  long r = 0;
  int  n = 0;

  /*
   * Read header
   */
  n = read( rd_socket, header, 30 );  /* 28 for header 2 for \r\n */

  if( n <= 0 || gwp_process_reply_header( ctrl, header, n ) < 0 ) {
    
    close( rd_socket );
    rd_socket = MINUSONE; 
    gwerror = GWE_DISCNCT;
    gw_logf( LOG_SERIOUS, "Unexpected disconnection from Goldwater Scheduler");
    free_cstring( &ctrl->cache );
    return NULL;
  }
  
  if( CSMax( ctrl->cache ) >= ctrl->total_size ) {
    reset_cstring( &ctrl->cache );
  } else {
    /* pre-size cache
     */
    free_cstring( &ctrl->cache );
    CreateCString( &ctrl->cache, ctrl->total_size );
  }

  /*
   * Read data straight into buffer
   */
  r = 0;
  do {
    n = read( rd_socket, CSStart(ctrl->cache) + r, CSMax(ctrl->cache) - r );
    if( n <= 0 ) {
      close( rd_socket );
      rd_socket = MINUSONE;
      gwerror = GWE_DISCNCT; 
      gw_logf( LOG_SERIOUS, 
	       "Unexpected disconnection from Goldwater Scheduler");
      free_cstring( &ctrl->cache );
      return NULL;
    }
    r += n;
  } while( r < ctrl->total_size );

  CSLen( ctrl->cache ) = ctrl->total_size;

  return CSStart( ctrl->cache );
}

/* ----------------------------------------------------------------------- */
/* end */
