/*
 * Goldwater Domain NetServer Reply Handler
 *
 * Copyright 2002 netFluid Technology Ltd
 *
 * Goldwater is written and maintained by netFluid Technology Ltd.
 * http://www.nfluid.co.uk
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * --------------------------------------------------------------------------
 * $Id: nsreply.c,v 1.2 2003/03/29 12:21:11 csmith Exp $
 * --------------------------------------------------------------------------
 */
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <fcntl.h>
#include <sys/mman.h>

#include <phlib/xl_types.h>
#include <phlib/xl_string.h>
#include <phlib/xl_memlst.h>
#include <phlib/xl_memory.h>
#include <phlib/xl_malloc.h>

#define GW_INTERNAL_FLAGS
#include <gwmsg.h>
#include <gwneng.h>
#include <gwnevent.h>
#include <gwxlapi.h>
#include <goldreq.h>
#include <bbquery.h>
#include <gwgbprm.h>
#include <gwlog.h>
#include <gwdomain.h>

#include <dbcom.h>

#include "nsctrl.h"
#include "nsreply.h"
#include "nsstate.h"
#include "version.h"

/* Prototype functions 
 */

/* ----------------------------------------------------------------------- */
/*
 * Sends request to service via message queue.
 * - Request may contain either the full data set or a spoolfile filename
 *
 * Return:
 *  1 = Successfully sent to service
 *  0 = Blocked, will try again (via events)
 * -1 = Failure to send to service
 */
int
send_reply_to_client( nsSession *sc, long len )
{
  static pid_t mypid = -1;

  if( mypid < 0 ) mypid = getpid();
  
  sc->req_msgf.ident       = 1;
  sc->req_msgf.dsize       = len;
  sc->req_msgf.flags       = sc->ctrl.flags;
  sc->req_msgf.total_size  = sc->ctrl.total_size;   /* Total data len */
  sc->req_msgf.dest_srv_id = sc->ctrl.destsrvid;
  sc->req_msgf.dest_dom_id = sc->fromdom_id;        /* ORIGINATORS DOMAIN ID */
  sc->req_msgf.rep_msgqid  = -1;
  sc->req_msgf.msgtype     = GWM_SVC_REPLY;
  uuid_copy( sc->req_msgf.msgid, sc->ctrl.msgid );
  sc->req_msgf.clid        = mypid;
  
  /* Service (or alias) name */
  strncpy( sc->req_msgf.service, sc->svc_info.service, SVC_NM_MAX );
  sc->req_msgf.service[ SVC_NM_MAX ] = '\0';
  
  /* Function name */
  sc->req_msgf.function[0] = '\0';

  return resend_reply_to_client( sc );
}

/* ----------------------------------------------------------------------- */
/*
 * Return:
 *  1 = Successfully sent to service
 *  0 = Blocked, will try again (via events)
 * -1 = Failure to send to service
 */
int
resend_reply_to_client( nsSession *sc )
{
  long mlen       = 0;
  
  mlen = ((long)&sc->req_msgf.data - (long)&sc->req_msgf.flags) + sc->req_msgf.dsize;
  
  if( msgsnd( sc->ctrl.org_r_msgqid,
	      (unsigned char*)&(sc->req_msgf), mlen, IPC_NOWAIT ) < 0 ) {
    
    if( errno == EAGAIN ) {
      /* msgsnd would have blocked..... Register event to re-send
       */
      xlapi_register_event( EV_USER_EVENT, EU_REPMSGSND_EAGAIN );
      return 0;
    }

    gw_logf( LOG_SERIOUS, "msgsnd failure (error %ld)", (long)errno );
    send_ack( sc, GWDomFailure, GWE_SYSFAIL );
    return -1;
  }

  send_ack( sc, GWSuccess, 0 );
  return 1;
}

/* ----------------------------------------------------------------------- */
/* end */
